# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
SAOCOM products.
"""
import logging
import re
import zipfile
from datetime import datetime
from enum import unique
from typing import Union

import geopandas as gpd
from lxml import etree
from sertit import path, rasters, vectors
from sertit.misc import ListEnum
from shapely.geometry import Polygon

from eoreader import DATETIME_FMT, EOREADER_NAME, cache
from eoreader.bands import SarBandNames
from eoreader.exceptions import InvalidProductError, InvalidTypeError
from eoreader.products import SarProduct, SarProductType
from eoreader.products.product import OrbitDirection
from eoreader.products.sar.sar_product import _ExtendedFormatter
from eoreader.utils import simplify

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class SaocomProductType(ListEnum):
    """
    SAOCOM Processing levels.
    Take a look
    `here <https://earth.esa.int/eogateway/catalog/saocom-data-products>`_
    """

    SLC = "L1A"
    """L1A - SLC (single look complex, slant range)"""

    DI = "L1B"
    """L1B - DI (detected image, ground range)"""

    GEC = "L1C"
    """L1C - GEC (geocoded on ellipsoid)"""

    GTC = "L1D"
    """L1D - GTC (geocoded on DEM)"""


@unique
class SaocomSensorMode(ListEnum):
    """
    SAOCOM Acquisition Mode mode.
    Take a look
    `here <https://saocom.veng.com.ar/L1-product-format-EN.pdf>`_
    """

    SM = "STRIPMAP"
    """Stripmap"""

    TN = "TOPSAR Narrow"
    """TOPSAR Narrow"""

    TW = "TOPSAR Wide"
    """TOPSAR Wide"""


@unique
class SaocomPolarization(ListEnum):
    """
    SAOCOM polarization mode.
    Take a look
    `here <https://saocom.veng.com.ar/L1-product-format-EN.pdf>`_

    acquiredPols:
    - HH
    - HV
    - VH
    - VV
    - HH-HV
    - VH-VV
    - HH-HV-VH-VV
    - LeftH-LeftV
    - RightH-RightV

    Polarization
    - HH
    - VV
    - HV
    - VH
    - CL/H
    - CL/V
    - CR/H
    - CR/V
    """

    SP = "SP"
    """"Single Polarization (HH or VV)"""

    DP = "DP"
    """"Dual Polarization (HH/HV or VV/VH)"""

    QP = "QP"
    """"Quadruple Polarization (HH/HV/VH/VV)"""

    CP = "CP"
    """"Compact Polarization (LH/LV or RH/RV): Not yet available"""


class SaocomProduct(SarProduct):
    """Class for SAOCOM-1 Products"""

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        See here
        `here <https://saocom.veng.com.ar/L1-product-format-EN.pdf>`_
        for more information (Beam Modes)

        .. WARNING::
            For SSC data:
                - We assume being in High Resolution (SE)
                - Incidence angle: we consider the worst option, around 20 degrees
        """
        # Find the polarization
        root, _ = self.read_mtd()

        # Open identifier
        polarization = SaocomPolarization.from_value(root.findtext(".//polMode"))
        if not polarization:
            raise InvalidProductError("polMode not found in metadata!")

        # For complex data, set regular ground range pixel_size and resolution provided by the constructor
        if self.sensor_mode == SaocomSensorMode.SM:
            def_res = 10.0
        elif self.sensor_mode == SaocomSensorMode.TN:
            if polarization == SaocomPolarization.QP:
                def_res = 50.0
            else:
                def_res = 30.0
        elif self.sensor_mode == SaocomSensorMode.TW:
            if polarization == SaocomPolarization.QP:
                def_res = 100.0
            else:
                def_res = 50.0

        else:
            raise InvalidProductError(f"Unknown sensor mode: {self.sensor_mode}")

        # Impossible to find anywhere the recommaneded pixel spacing for SAOCOM data
        # SLC: For STRIPMAP data this product is sampled at the natural pixel spacing
        # Decision : For other SAR images, usually: pixel_size = resolution / 2.
        # TODO: Is it the case here ?
        self.pixel_size = def_res / 2.0
        self.resolution = def_res

    def _set_instrument(self) -> None:
        """
        Set instrument

        SAOCOM: https://earth.esa.int/eogateway/missions/saocom
        """
        self.instrument = "SAR L-band"

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        # Private attributes
        self._raw_band_regex = "*Data*-{!l}*"
        self._band_folder = self.path.joinpath("Data")

        # SNAP cannot process its archive
        self.needs_extraction = True

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting product-type, band names and so on)
        """
        self.snap_filename = f"{self.name}.xemt"

        # Post init done by the super class
        super()._post_init(**kwargs)

    @cache
    def wgs84_extent(self) -> gpd.GeoDataFrame:
        """
        Get the WGS84 extent of the file before any reprojection.
        This is useful when the SAR pre-process has not been done yet.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"TSX1_SAR__MGD_SE___SM_S_SRA_20160229T223018_20160229T223023"
            >>> prod = Reader().open(path)
            >>> prod.wgs84_extent()
                                                        geometry
            0  POLYGON ((106.65491 -6.39693, 106.96233 -6.396...

        Returns:
            gpd.GeoDataFrame: WGS84 extent as a gpd.GeoDataFrame

        """
        root, _ = self.read_mtd()

        # Compute extent corners
        corners = [
            [float(vertex.findtext("lon")), float(vertex.findtext("lat"))]
            for vertex in root.iterfind(".//frame/vertex")
        ]
        # TODO: ensure that the polygon is valid ?

        return gpd.GeoDataFrame(
            geometry=[Polygon(corners)],
            crs=vectors.WGS84,
        )

        # # Open extent KML file
        # if self.product_type == SaocomProductType.SLC:
        #     pass
        #
        # else:
        #     try:
        #         extent_file = next(self.path.glob("**/Images/*.kml"))
        #     except IndexError as ex:
        #         raise InvalidProductError(
        #             f"Extent file (products.kml) not found in {self.path}"
        #         ) from ex
        #
        #     extent_wgs84 = vectors.read(extent_file).envelope
        #
        #     return gpd.GeoDataFrame(geometry=extent_wgs84.geometry, crs=extent_wgs84.crs)

    @cache
    @simplify
    def footprint(self) -> gpd.GeoDataFrame:
        """
        Get UTM footprint of the products (without nodata, *in french == emprise utile*)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.footprint()
               index                                           geometry
            0      0  POLYGON ((199980.000 4500000.000, 199980.000 4...

        Returns:
            gpd.GeoDataFrame: Footprint as a GeoDataFrame
        """
        # WARNING: Sometimes the product seems to contain several tiles that are not contiguous
        # Do not simplify geometry then
        return rasters.get_valid_vector(
            self.get_default_band_path()
        )  # Processed by SNAP: the nodata is set

    def get_raw_band_paths(self, **kwargs) -> dict:
        """
        Return the existing band paths (as they come with the archived products).

        Args:
            **kwargs: Additional arguments

        Returns:
            dict: Dictionary containing the path of every band existing in the raw products
        """
        extended_fmt = _ExtendedFormatter()
        cuss_file = next(self.path.glob("*.zip"))
        band_paths = {}
        for band in SarBandNames.speckle_list():
            band_regex = extended_fmt.format(self._raw_band_regex, band.value)

            try:
                # Get as a list but keep only the first item (SLC with multiple swaths)
                raw_paths = path.get_archived_rio_path(
                    cuss_file, band_regex.replace("*", ".*"), as_list=True
                )

                # Remove .xml files and keep only the first item
                band_paths[band] = [p for p in raw_paths if not p.endswith(".xml")][0]
            except FileNotFoundError:
                continue

        return band_paths

    def _set_product_type(self) -> None:
        """Set products type"""
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open identifier
        prod_type = root.findtext(".//procLevel")
        if not prod_type:
            raise InvalidProductError("procLevel not found in metadata!")

        self.product_type = SaocomProductType.from_value(prod_type)

        if self.product_type in [
            SaocomProductType.DI,
            SaocomProductType.GEC,
            SaocomProductType.GTC,
        ]:
            self.sar_prod_type = SarProductType.GDRG
        elif self.product_type == SaocomProductType.SLC:
            self.sar_prod_type = SarProductType.CPLX
        else:
            raise NotImplementedError(
                f"{self.product_type.value} product type is not available for {self.name}"
            )
        if self.product_type == SaocomProductType.DI:
            LOGGER.warning(
                "DI (Detected Image) product type has never been tested for %s data. "
                "Use it at your own risks !",
                self.constellation.value,
            )

    def _set_sensor_mode(self) -> None:
        """
        Get products type from TerraSAR-X products name (could check the metadata too)
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open identifier
        imaging_mode = root.findtext(".//acqMode")
        if not imaging_mode:
            raise InvalidProductError("acqMode not found in metadata!")

        # Get sensor mode
        try:
            self.sensor_mode = getattr(SaocomSensorMode, imaging_mode)
        except ValueError as ex:
            raise InvalidTypeError(f"Invalid sensor mode for {self.name}") from ex

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"TSX1_SAR__MGD_SE___SM_S_SRA_20160229T223018_20160229T223023"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2016, 2, 29, 22, 30, 18)
            >>> prod.get_datetime(as_datetime=False)
            '20160229T223018'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        if self.datetime is None:
            # Get MTD XML file
            root, _ = self.read_mtd()

            # Open identifier
            acq_date = root.findtext(".//acquisitionTime/startTime")
            if not acq_date:
                raise InvalidProductError(
                    "acquisitionTime/startTime not found in metadata!"
                )

            # Convert to datetime
            date = datetime.strptime(acq_date, "%Y-%m-%dT%H:%M:%S.%f")
        else:
            date = self.datetime

        if not as_datetime:
            date = date.strftime(DATETIME_FMT)

        return date

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open identifier
        try:
            name = path.get_filename(root.findtext(".//dataFile/componentPath"))
        except TypeError:
            raise InvalidProductError("dataFile/componentPath not found in metadata!")

        return name

    @cache
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"TSX1_SAR__MGD_SE___SM_S_SRA_20200605T042203_20200605T042211"
            >>> prod = Reader().open(path)
            >>> prod.read_mtd()
            (<Element level1Product at 0x1b845b7ab88>, {})

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces
        """
        mtd_from_path = "xemt"

        return self._read_mtd_xml(mtd_from_path)

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing.

        Returns:
            str: Quicklook path
        """
        quicklook_path = None
        try:
            qlk_path, qlk_exists = self._get_out_path(f"{self.condensed_name}_QLK.tif")

            if not qlk_exists:
                try:
                    zip_path = next(self.path.glob(f"{self.name}.zip"))

                    with zipfile.ZipFile(zip_path) as zip_ds:
                        filenames = [f.filename for f in zip_ds.filelist]
                        regex = re.compile("Images/.*png")
                        quicklook_path = zip_ds.extract(
                            list(filter(regex.match, filenames))[0], qlk_path
                        )
                except FileNotFoundError:
                    quicklook_path = next(self.path.glob("Images/*.png"))
                quicklook_path = str(quicklook_path)
        except (StopIteration, FileNotFoundError):
            LOGGER.warning(f"No quicklook found in {self.condensed_name}")

        return quicklook_path

    @cache
    def get_orbit_direction(self) -> OrbitDirection:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_orbit_direction().value
            "DESCENDING"

        Returns:
            OrbitDirection: Orbit direction (ASCENDING/DESCENDING)
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Get the orbit direction
        try:
            od = OrbitDirection.from_value(root.findtext(".//OrbitDirection"))

        except TypeError:
            raise InvalidProductError("orbitDirection not found in metadata!")

        return od



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Sentinel-2 MPC products """
import difflib
import logging

import numpy as np
import planetary_computer
import xarray as xr
from lxml import etree
from pystac import Item
from sertit import AnyPath
from sertit.types import AnyPathStrType, AnyPathType

from eoreader import EOREADER_NAME
from eoreader.bands import BandNames
from eoreader.exceptions import InvalidProductError
from eoreader.products import S2E84Product
from eoreader.products.optical.optical_product import RawUnits
from eoreader.products.stac_product import StacProduct
from eoreader.reader import Constellation

LOGGER = logging.getLogger(EOREADER_NAME)


class S2MpcStacProduct(StacProduct, S2E84Product):
    def __init__(
        self,
        product_path: AnyPathStrType = None,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        remove_tmp: bool = False,
        **kwargs,
    ) -> None:
        self.kwargs = kwargs
        """Custom kwargs"""

        self._processing_baseline = None

        # Copy the kwargs
        super_kwargs = kwargs.copy()

        # Get STAC Item
        self.item = None
        """ STAC Item of the product """
        self.item = super_kwargs.pop("item", None)
        if self.item is None:
            try:
                self.item = Item.from_file(product_path)
            except TypeError:
                raise InvalidProductError(
                    "You should either fill 'product_path' or 'item'."
                )

        self.default_clients = [planetary_computer]
        self.clients = super_kwargs.pop("client", self.default_clients)

        if product_path is None:
            # Canonical link is always the second one
            # TODO: check if ok
            product_path = AnyPath(self.item.links[1].target).parent

        # Initialization from the super class
        super().__init__(product_path, archive_path, output_path, remove_tmp, **kwargs)

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        self._raw_units = RawUnits.REFL
        self._has_cloud_cover = True
        self._use_filename = False
        self.needs_extraction = False

        self.stac_mtd = self.item.to_dict()

        # Pre init done by the super class
        super(S2E84Product, self)._pre_init(**kwargs)

    def _get_constellation(self) -> Constellation:
        """Getter of the constellation: force S2."""
        return Constellation.S2

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting sensor type, band names and so on)
        """
        # Get processing baseline: N0213 -> 02.13
        pr_baseline = float(self.split_name[3][1:]) / 100
        self._processing_baseline = pr_baseline

        # Pre init done by the super class
        super(S2E84Product, self)._post_init(**kwargs)

    def _get_name(self) -> str:
        """
        Set product real name.

        Returns:
            str: True name of the product (from metadata)
        """
        return self.item.properties["s2:product_uri"]

    def _to_reflectance(
        self,
        band_arr: xr.DataArray,
        band_path: AnyPathType,
        band: BandNames,
        **kwargs,
    ) -> xr.DataArray:
        """
        Converts band to reflectance

        Args:
            band_arr (xr.DataArray): Band array to convert
            band_path (AnyPathType): Band path
            band (BandNames): Band to read
            **kwargs: Other keywords

        Returns:
            xr.DataArray: Band in reflectance
        """
        # TODO: use mtd for offset and quantif value
        if self._processing_baseline < 4.0:
            offset = 0.0
        else:
            offset = -1000.0
        quantif_value = 10000.0

        # Compute the correct radiometry of the band
        band_arr = (band_arr + offset) / quantif_value

        return band_arr.astype(np.float32)

    def _get_path(self, file_id: str, ext="tif") -> str:
        """
        Get either the archived path of the normal path of a tif file

        Args:
            band_id (str): Band ID

        Returns:
            AnyPathType: band path
        """
        if file_id.lower() in self.item.assets:
            asset_name = file_id.lower()
        elif file_id in [band.id for band in self.bands.values() if band is not None]:
            asset_name = [
                band.name
                for band in self.bands.values()
                if band is not None and f"{band.id}" == file_id
            ][0]
        else:
            try:
                asset_name = difflib.get_close_matches(
                    file_id, self.item.assets.keys(), cutoff=0.5, n=1
                )[0]
            except Exception:
                raise FileNotFoundError(
                    f"Impossible to find an asset in {list(self.item.assets.keys())} close enough to '{file_id}'"
                )

        return self.sign_url(self.item.assets[asset_name].href)

    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read mtd.

        Args:
            force_pd (bool): If collection 2, return a pandas.DataFrame instead of an XML root + namespace

        Returns:
            Tuple[Union[pd.DataFrame, etree._Element], dict]:
                Metadata as a Pandas.DataFrame or as (etree._Element, dict): Metadata XML root and its namespaces
        """
        return self._read_mtd_xml_stac(self._get_path("granule-metadata"))

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing.

        Returns:
            str: Quicklook path
        """
        return self._get_path("rendered_preview")



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Super-class of Planet products.
See
`Product specs <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_
for more information.
"""
import logging
from abc import abstractmethod
from enum import unique
from typing import Tuple, Union

import geopandas as gpd
import numpy as np
import rasterio
import xarray as xr
from lxml import etree
from rasterio.enums import Resampling
from sertit import path, rasters, strings
from sertit.misc import ListEnum
from sertit.types import AnyPathStrType, AnyPathType

from eoreader import EOREADER_NAME, cache, utils
from eoreader.bands import (
    ALL_CLOUDS,
    BLUE,
    CA,
    CIRRUS,
    CLOUDS,
    GREEN,
    GREEN_1,
    NARROW_NIR,
    NIR,
    RAW_CLOUDS,
    RED,
    SHADOWS,
    VRE_1,
    YELLOW,
    BandNames,
    to_str,
)
from eoreader.exceptions import InvalidProductError, InvalidTypeError
from eoreader.products import OpticalProduct, OrbitDirection
from eoreader.products.optical.optical_product import RawUnits
from eoreader.reader import Constellation
from eoreader.utils import simplify

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class PlanetMaskType(ListEnum):
    """
    Planet Mask Type
    UDM2 > UDM > NONE
    """

    UDM2 = "Usable Data Mask"
    """
    The usable data mask file provides information on areas of usable data within an image (e.g. clear, snow, shadow, light haze, heavy haze and cloud).
    The pixel size after orthorectification will be 3.125 m for PlanetScope OrthoTiles and 3.0m for PlanetScope Scenes.
    The usable data mask is a raster image having the same dimensions as the image product, comprised of 8 bands, where each band represents a specific usability class mask.
    The usability masks are mutually exclusive, and a value of one indicates that the pixel is assigned to that usability class.
    - Band 1: clear mask (a value of “1” indicates the pixel is clear, a value of “0” indicates that the pixel is not clear and is one of the 5 remaining classes below)
    - Band 2: snow mask
    - Band 3: shadow mask
    - Band 4: light haze mask
    - Band 5: heavy haze mask
    - Band 6: cloud mask
    - Band 7: confidence map (a value of “0” indicates a low confidence in the assigned classification, a value of “100” indicates a high confidence in the assigned classification)
    - Band 8: unusable data mask
    """

    UDM = "Unusable Data Mask"
    """
    The unusable data mask file provides information on areas of unusable data within an image (e.g. cloud and non-imaged areas).
    The pixel size after orthorectification will be 3.125 m for PlanetScope OrthoTiles, 3.0m for PlanetScope Scenes, 50m for RapidEye, and 0.8 m for SkySat.
    It is suggested that when using the file to check for usable data, a buffer of at least 1 pixel should be considered.
    Each bit in the 8-bit pixel identifies whether the corresponding part of the product contains useful imagery:
    - Bit 0: Identifies whether the area contains blackfill in all bands (this area was not imaged). A value of “1” indicates blackfill.
    - Bit 1: Identifies whether the area is cloud covered. A value of “1” indicates cloud coverage.
    Cloud detection is performed on a decimated version of the image (i.e. the browse image) and hence small clouds may be missed.
    Cloud areas are those that have pixel values in the assessed band (Red, NIR or Green) that are above a configurable threshold.
    This algorithm will:
    - Assess snow as cloud
    - Assess cloud shadow as cloud free
    - Assess haze as cloud free
    - Bit 2: Identifies whether the area contains missing (lost during downlink) or suspect (contains down-link errors) data in band 1.
    A value of “1” indicates missing/suspect data. If the product does not include this band, the value is set to “0”.
    - Bit 3: Identifies whether the area contains missing (lost during downlink and hence blackfilled) or suspect (contains downlink errors) data in the band 2.
    A value of “1” indicates missing/suspect data. If the product does not include this band, the value is set to “0”.
    - Bit 4: Identifies whether the area contains missing (lost during downlink) or suspect (contains downlink errors) data in the band 3.
    A value of “1” indicates missing/suspect data. If the product does not include this band, the value is set to “0”.
    - Bit 5: Identifies whether the area contains missing (lost during downlink) or suspect (contains downlink errors) data in band 4.
    A value of “1” indicates missing/suspect data. If the product does not include this band, the value is set to “0”.
    - Bit 6: Identifies whether the area contains missing (lost during downlink) or suspect (contains downlink errors) data in band 5.
    A value of “1” indicates missing/suspect data. If the product does not include this band, the value is set to “0”.
    - Bit 7: Is currently set to “0”.

    The UDM information is found in band 8 of the Usable Data Mask file.
    """

    NONE = "None"
    """
    The product has no mask. Only valid for some RapidEye old products
    """


class PlanetProduct(OpticalProduct):
    """
    Super-class of Planet products.
    See
    `Earth Online <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_
    for more information.

    The scaling factor to retrieve the calibrated radiance is 0.01.
    """

    def __init__(
        self,
        product_path: AnyPathStrType,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        remove_tmp: bool = False,
        **kwargs,
    ) -> None:
        self._nsmap_key = None
        self._merged = False
        self._to_merge = False

        # Initialization from the super class
        super().__init__(product_path, archive_path, output_path, remove_tmp, **kwargs)

        if len(self._get_stack_path(as_list=True)) > 1:
            # self._to_merge = True
            self._merge_subdatasets_mtd()

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        # Update namespace map key (if needed)
        if self.constellation == Constellation.RE:
            self._nsmap_key = "re"
        elif self.constellation == Constellation.PLA:
            self._nsmap_key = "ps"

        # Tilename
        if self.constellation in [Constellation.RE, Constellation.PLA]:
            # Get MTD XML file
            root, nsmap = self.read_mtd()

            # Manage constellation
            self.tile_name = root.findtext(f".//{nsmap[self._nsmap_key]}tileId")

        self.needs_extraction = False

        # Manage mask type
        try:
            if self.is_archived:
                path.get_archived_path(self.path, r".*udm2.*\.tif")
            else:
                next(self.path.glob("**/*udm2*.tif"))
            self._mask_type = PlanetMaskType.UDM2
        except (FileNotFoundError, StopIteration):
            try:
                if self.is_archived:
                    path.get_archived_path(self.path, r".*udm.*\.tif")
                else:
                    next(self.path.glob("**/*udm*.tif"))
                self._mask_type = PlanetMaskType.UDM
            except (FileNotFoundError, StopIteration):
                LOGGER.warning(
                    "UDM mask not found. This product won't be cleaned and won't have any cloud band."
                )
                self._mask_type = PlanetMaskType.NONE

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting sensor type, band names and so on)
        """
        # Manage Raw unit
        band_name = (
            path.get_filename(self._get_stack_path(as_list=False)).upper().split("_")
        )
        if "SR" in band_name:
            self._raw_units = RawUnits.REFL
        elif "DN" in band_name:
            # Seems only for basic products
            self._raw_units = RawUnits.DN
        elif "VISUAL" in band_name:
            # Seems only for basic products
            self._raw_units = RawUnits.NONE
        else:
            # If not specified, Planet product are in scaled radiance (*0.01)
            self._raw_units = RawUnits.RAD

        # Post init done by the super class
        super()._post_init(**kwargs)

    @abstractmethod
    def _get_stack_path(self, as_list: bool = False) -> Union[str, list]:
        """
        Get Planet stack path(s)

        Args:
            as_list (bool): Get stack path as a list (useful if several subdatasets are present)

        Returns:
            Union[str, list]: Stack path(s)
        """
        raise NotImplementedError

    def _get_udm_path(self, as_list: bool = False) -> Union[str, list]:
        """
        Get Planet UDM path

        Args:
            as_list (bool): Get stack path as a list (useful if several subdatasets are present)

        Returns:
            Union[str, list]: Stack path(s)
        """
        if self._merged:
            udm_path, _ = self._get_out_path(f"{self.condensed_name}_udm.vrt")
            if as_list:
                udm_path = [udm_path]
        else:
            udm_path = self._get_path(
                "udm", "tif", invalid_lookahead="udm2", as_list=as_list
            )

        return udm_path

    def _get_udm2_path(self, as_list: bool = False) -> Union[str, list]:
        """
        Get Planet UDM2 path

        Args:
            as_list (bool): Get stack path as a list (useful if several subdatasets are present)

        Returns:
            Union[str, list]: Stack path(s)
        """
        if self._merged:
            udm2_path, _ = self._get_out_path(f"{self.condensed_name}_udm2.vrt")
            if as_list:
                udm2_path = [udm2_path]
        else:
            udm2_path = self._get_path("udm2", "tif", as_list=as_list)

        return udm2_path

    def _merge_subdatasets(self) -> Tuple[AnyPathType, bool]:
        """
        Merge subdataset, when several Planet products avec been ordered together
        Will create a reflectance (if possible) VRT, a UDM/UDM2 VRT.

        Returns:
            Tuple[AnyPathType, bool]: Analytic VRT and if already existing
        """
        if self.is_archived:
            # VRT cannot be created from inside a ZIP
            raise InvalidProductError(
                "EOReader doens't handle archived Planet data with multiple subdataset. Please extract this product."
            )
            # TODO: merge_geotiff BUT handle reflectances for every subdataset!
            # Relevant ? Maybe not as it takes would a lot of time to merge

        if path.is_cloud_path(self.path):
            # VRT cannot be created from data stored in the cloud
            raise InvalidProductError(
                "EOReader doens't handle cloud-stored Planet data with multiple subdataset. Please download this product on disk."
            )
            # Relevant ? Maybe not as it takes would a lot of time to download, or a lot of time to merge as geotiffs

        analytic_vrt_path, analytic_vrt_exists = self._get_out_path(
            f"{self.condensed_name}_analytic.vrt"
        )
        if not analytic_vrt_exists:
            LOGGER.debug("Creating raster VRT")
            rasters.merge_vrt(
                self._get_stack_path(as_list=True), analytic_vrt_path, abs_path=True
            )

        udm_vrt_path, udm_vrt_exists = self._get_out_path(
            f"{self.condensed_name}_udm.vrt"
        )
        if not udm_vrt_exists:
            udm_paths = self._get_udm_path(as_list=True)
            if udm_paths:
                LOGGER.debug("Creating UDM VRT")
                nodata = "1"
                rasters.merge_vrt(
                    self._get_udm_path(as_list=True),
                    udm_vrt_path,
                    abs_path=True,
                    **{"-srcnodata": nodata, "-vrtnodata": nodata},
                )

        udm2_vrt_path, udm2_vrt_exists = self._get_out_path(
            f"{self.condensed_name}_udm2.vrt"
        )
        if not udm2_vrt_exists:
            LOGGER.debug("Creating UDM2 VRT")
            # Nodata values
            # see: https://developers.planet.com/docs/data/udm-2/#udm2-bands
            # Band 1    Clear map       [0, 1] 	    0: not clear, 1: clear
            # Band 2 	Snow map 	    [0, 1] 	    0: no snow or ice, 1: snow or ice
            # Band 3 	Shadow map 	    [0, 1] 	    0: no shadow, 1: shadow
            # Band 4 	Light haze map 	[0, 1] 	    0: no light haze, 1: light haze
            # Band 5 	Heavy haze map 	[0, 1] 	    0: no heavy haze, 1: heavy haze
            # Band 6 	Cloud map 	    [0, 1] 	    0: no cloud, 1: cloud
            # Band 7 	Confidence map 	[ 0 - 100] 	percentage value: per-pixel algorithmic confidence in classification
            # Band 8 	Unusable pixels  --         Equivalent to the UDM asset: a value of “1” indicates blackfill.
            nodata = strings.to_cmd_string("0 0 0 0 0 0 0 1")
            rasters.merge_vrt(
                self._get_udm2_path(as_list=True),
                udm2_vrt_path,
                abs_path=True,
                **{"-srcnodata": nodata, "-vrtnodata": nodata},
            )

        self._merged = True

        return analytic_vrt_path, analytic_vrt_exists

    @abstractmethod
    def _merge_subdatasets_mtd(self) -> None:
        """
        Merge subdataset, when several Planet products avec been ordered together
        Will create a reflectance (if possible) VRT, a UDM/UDM2 VRT and a merged metadata XML file.

        Constellation specific
        """
        raise NotImplementedError

    @cache
    @simplify
    def footprint(self) -> gpd.GeoDataFrame:
        """
        Get real footprint of the products (without nodata, in french == emprise utile)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"LC08_L1GT_023030_20200518_20200527_01_T2"
            >>> prod = Reader().open(path)
            >>> prod.footprint()
               index                                           geometry
            0      0  POLYGON ((366165.000 4899735.000, 366165.000 4...

        Overload of the generic function because landsat nodata seems to be different in QA than in regular bands.
        Indeed, nodata pixels vary according to the band sensor footprint,
        whereas QA nodata is where at least one band has nodata.

        We chose to keep QA nodata values for the footprint in order to show where all bands are valid.

        **TL;DR: We use the QA nodata value to determine the product's footprint**.

        Returns:
            gpd.GeoDataFrame: Footprint as a GeoDataFrame
        """
        nodata = self._load_nodata()

        # Vectorize the nodata band
        # Don't use convex hull as the product can be cropped to an AOI!
        footprint = rasters.vectorize(
            nodata, values=1, keep_values=False, dissolve=True
        )

        return gpd.GeoDataFrame(geometry=footprint.geometry, crs=footprint.crs)

    def get_band_paths(
        self, band_list: list, pixel_size: float = None, **kwargs
    ) -> dict:
        """
        Return the paths of required bands.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> from eoreader.bands import *
            >>> path = r"SENTINEL2A_20190625-105728-756_L2A_T31UEQ_C_V2-2"
            >>> prod = Reader().open(path)
            >>> prod.get_band_paths([GREEN, RED])
            {
                <SpectralBandNames.GREEN: 'GREEN'>:
                'SENTINEL2A_20190625-105728-756_L2A_T31UEQ_C_V2-2/SENTINEL2A_20190625-105728-756_L2A_T31UEQ_C_V2-2_FRE_B3.tif',
                <SpectralBandNames.RED: 'RED'>:
                'SENTINEL2A_20190625-105728-756_L2A_T31UEQ_C_V2-2/SENTINEL2A_20190625-105728-756_L2A_T31UEQ_C_V2-2_FRE_B4.tif'
            }

        Args:
            band_list (list): List of the wanted bands
            pixel_size (float): Band pixel size
            kwargs: Other arguments used to load bands

        Returns:
            dict: Dictionary containing the path of each queried band
        """
        band_paths = {}
        band_path = self._get_stack_path(as_list=False)
        for band in band_list:
            # Get clean band path
            clean_band = self._get_clean_band_path(
                band, pixel_size=pixel_size, **kwargs
            )
            if clean_band.is_file():
                band_paths[band] = clean_band
            else:
                band_paths[band] = band_path

        return band_paths

    def _read_band(
        self,
        band_path: AnyPathType,
        band: BandNames = None,
        pixel_size: Union[tuple, list, float] = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> xr.DataArray:
        """
        Read band from disk.

        .. WARNING::
            Invalid pixels are not managed here

        Args:
            band_path (AnyPathType): Band path
            band (BandNames): Band to read
            pixel_size (Union[tuple, list, float]): Size of the pixels of the wanted band, in dataset unit (X, Y)
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands
        Returns:
            xr.DataArray: Band xarray

        """
        with rasterio.open(str(band_path)) as dst:
            # Manage the case if we open a simple band (EOReader processed bands)
            if dst.count == 1:
                # Read band
                band_arr = utils.read(
                    band_path,
                    pixel_size=pixel_size,
                    size=size,
                    resampling=Resampling.bilinear,
                    **kwargs,
                )

            # Manage the case if we open a stack (native DIMAP bands)
            else:
                band_arr = utils.read(
                    band_path,
                    pixel_size=pixel_size,
                    size=size,
                    resampling=Resampling.bilinear,
                    indexes=[self.bands[band].id],
                    **kwargs,
                )

        # Pop useless long name
        if "long_name" in band_arr.attrs:
            band_arr.attrs.pop("long_name")

        # To float32
        if band_arr.dtype != np.float32:
            band_arr = band_arr.astype(np.float32)

        return band_arr

    def _manage_invalid_pixels(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage invalid pixels (Nodata, saturated, defective...)
        See
        `here <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_
        (unusable data mask) for more information.

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as a SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        # Nodata
        no_data_mask = self._load_nodata(
            size=(band_arr.rio.width, band_arr.rio.height), **kwargs
        ).values

        # Dubious pixels mapping
        # See: https://community.planet.com/planet-s-community-forum-3/planetscope-8-bands-and-udm-mask-245?postid=436#post436
        dubious_bands = {
            BLUE: 2,
            GREEN: 3,
            RED: 4,
            VRE_1: 5,
            NIR: 6,
            NARROW_NIR: 6,
            CA: 7,
            GREEN_1: 7,
            YELLOW: 7,
        }

        # Open unusable mask
        udm = self.open_mask(
            "UNUSABLE", size=(band_arr.rio.width, band_arr.rio.height), **kwargs
        )
        # Workaround:
        # FutureWarning: The :code:`numpy.expand_dims` function is not implemented by Dask array.
        # You may want to use the da.map_blocks function or something similar to silence this warning.
        # Your code may stop working in a future release.
        dubious_mask = rasters.read_bit_array(udm.values, dubious_bands[band])

        # Combine masks
        mask = no_data_mask | dubious_mask

        # -- Merge masks
        return self._set_nodata_mask(band_arr, mask)

    def _manage_nodata(
        self, band_arr: xr.DataArray, band: BandNames, **kwargs
    ) -> xr.DataArray:
        """
        Manage only nodata pixels

        Args:
            band_arr (xr.DataArray): Band array
            band (BandNames): Band name as a SpectralBandNames
            kwargs: Other arguments used to load bands

        Returns:
            xr.DataArray: Cleaned band array
        """
        # Nodata
        no_data_mask = self._load_nodata(
            size=(band_arr.rio.width, band_arr.rio.height), **kwargs
        ).values

        # -- Merge masks
        return self._set_nodata_mask(band_arr, no_data_mask)

    def _load_bands(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load bands as numpy arrays with the same pixel size (and same metadata).

        Args:
            bands list: List of the wanted bands
            pixel_size (float): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Other arguments used to load bands
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        # Return empty if no band are specified
        if not bands:
            return {}

        # Get band paths
        band_paths = self.get_band_paths(bands, pixel_size, **kwargs)

        # Open bands and get array (resampled if needed)
        band_arrays = self._open_bands(
            band_paths, pixel_size=pixel_size, size=size, **kwargs
        )

        return band_arrays

    def _has_cloud_band(self, band: BandNames) -> bool:
        """
        Does this product has the specified cloud band ?
        """
        # NOTE: CIRRUS == HEAVY HAZE

        # FROM DOCUMENTATION: https://developers.planet.com/docs/data/udm-2/
        # Percent of heavy haze values in dataset.
        # Heavy haze values represent scene content areas (non-blackfilled) that contain thin low altitude clouds,
        # higher altitude cirrus clouds, soot and dust which allow fair recognition of land cover features,
        # but not having reliable interpretation of the radiometry or surface reflectance.
        if self._mask_type == PlanetMaskType.UDM2:
            has_band = True
        elif self._mask_type == PlanetMaskType.UDM:
            if band in [SHADOWS, CIRRUS]:
                has_band = False
            else:
                has_band = True
        else:
            has_band = False

        return has_band

    def _open_clouds(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load cloud files as xarrays.

        CIRRUS is HEAVY_HAZE

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        if self._mask_type == PlanetMaskType.UDM2:
            return self._open_clouds_udm2(bands, pixel_size, size, **kwargs)
        else:
            # UDM
            return self._open_clouds_udm(bands, pixel_size, size, **kwargs)

    def _open_clouds_udm2(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load cloud files as xarrays.

        CIRRUS is HEAVY_HAZE

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        band_dict = {}

        # Load default xarray as a template
        def_xarr = self._read_band(
            self.get_default_band_path(),
            band=self.get_default_band(),
            pixel_size=pixel_size,
            size=size,
            **kwargs,
        )

        # Load nodata
        nodata = self._load_nodata(pixel_size, size, **kwargs).data

        if bands:
            for band in bands:
                if band == ALL_CLOUDS:
                    cloud = self._create_mask(
                        def_xarr.rename(ALL_CLOUDS.name),
                        (
                            self.open_mask("CLOUD", pixel_size, size, **kwargs).data
                            & self.open_mask("SHADOW", pixel_size, size, **kwargs).data
                            & self.open_mask(
                                "HEAVY_HAZE", pixel_size, size, **kwargs
                            ).data
                        ),
                        nodata,
                    )
                elif band == SHADOWS:
                    cloud = self._create_mask(
                        def_xarr.rename(SHADOWS.name),
                        self.open_mask("SHADOW", pixel_size, size, **kwargs).data,
                        nodata,
                    )
                elif band == CLOUDS:
                    cloud = self._create_mask(
                        def_xarr.rename(CLOUDS.name),
                        self.open_mask("CLOUD", pixel_size, size, **kwargs).data,
                        nodata,
                    )
                elif band == CIRRUS:
                    cloud = self._create_mask(
                        def_xarr.rename(CIRRUS.name),
                        self.open_mask("HEAVY_HAZE", pixel_size, size, **kwargs).data,
                        nodata,
                    )
                elif band == RAW_CLOUDS:
                    cloud = utils.read(
                        self._get_udm2_path(), pixel_size, size, **kwargs
                    )
                else:
                    raise InvalidTypeError(
                        f"Non existing cloud band for {self.constellation.value}: {band}"
                    )

                # Rename
                band_name = to_str(band)[0]

                # Multi bands -> do not change long name
                if band != RAW_CLOUDS:
                    cloud.attrs["long_name"] = band_name
                band_dict[band] = cloud.rename(band_name).astype(np.float32)

        return band_dict

    def _open_clouds_udm(
        self,
        bands: list,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> dict:
        """
        Load cloud files as xarrays.

        CIRRUS is HEAVY_HAZE

        Args:
            bands (list): List of the wanted bands
            pixel_size (int): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.
            kwargs: Additional arguments
        Returns:
            dict: Dictionary {band_name, band_xarray}
        """
        band_dict = {}

        # Load default xarray as a template
        def_xarr = self._read_band(
            self.get_default_band_path(),
            band=self.get_default_band(),
            pixel_size=pixel_size,
            size=size,
            **kwargs,
        )
        # Open mask (here we know we have a UDM file, as the product is supposed to have the band)
        udm = self.open_mask_udm(pixel_size, size, **kwargs)

        if bands:
            for band in bands:
                if band in [ALL_CLOUDS, CLOUDS]:
                    # Load nodata
                    nodata, clouds = rasters.read_bit_array(udm.compute(), [0, 1])

                    cloud = self._create_mask(
                        def_xarr.rename(ALL_CLOUDS.name),
                        clouds,
                        nodata,
                    )
                elif band == RAW_CLOUDS:
                    cloud = udm
                else:
                    raise InvalidTypeError(
                        f"Non existing cloud band for {self.constellation.value}: {band}"
                    )

                # Rename
                band_name = to_str(band)[0]

                # Multi bands -> do not change long name
                if band != RAW_CLOUDS:
                    cloud.attrs["long_name"] = band_name
                band_dict[band] = cloud.rename(band_name).astype(np.float32)

        return band_dict

    def open_mask(
        self,
        mask_id: str,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> Union[xr.DataArray, None]:
        """
        Open a Planet UDM2 (Usable Data Mask) mask, band by band, as a xarray.
        Returns None if the mask is not available.

        Do not open cloud mask with this function. Use :code:`load` instead.

        See `UDM2 specifications <https://developers.planet.com/docs/data/udm-2/>`_ for more
        information.

        Accepted mask IDs:

        - :code:`CLEAR`:      Band 1     Clear map        [0, 1]  0: not clear, 1: clear
        - :code:`SNOW`:       Band 2     Snow map         [0, 1]  0: no snow or ice, 1: snow or ice
        - :code:`SHADOW`:     Band 3     Shadow map       [0, 1]  0: no shadow, 1: shadow
        - :code:`LIGHT_HAZE`: Band 4     Light haze map   [0, 1]  0: no light haze, 1: light haze
        - :code:`HEAVY_HAZE`: Band 5     Heavy haze map   [0, 1]  0: no heavy haze, 1: heavy haze
        - :code:`CLOUD`:      Band 6     Cloud map        [0, 1]  0: no cloud, 1: cloud
        - :code:`CONFIDENCE`: Band 7     Confidence map   [0-100] %age value: per-pixel algorithmic confidence in classif
        - :code:`UNUSABLE`:   Band 8     Unusable pixels  --      Equivalent to the UDM asset

        .. code-block:: python

            >>> from eoreader.bands import *
            >>> from eoreader.reader import Reader
            >>> path = r"SENTINEL2B_20190401-105726-885_L2A_T31UEQ_D_V2-0.zip"
            >>> prod = Reader().open(path)
            >>> prod.open_mask("EDG", GREEN)
            array([[[0, ..., 0]]], dtype=uint8)

        Args:
            mask_id: Mask ID
            pixel_size (float): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.

        Returns:
            Union[xarray.DataArray, None]: Mask array

        """
        if self._mask_type == PlanetMaskType.UDM2:
            mask = self.open_mask_udm2(mask_id, pixel_size, size, **kwargs)
        elif self._mask_type == PlanetMaskType.UDM:
            mask = self.open_mask_udm(pixel_size, size, **kwargs)
        else:
            def_xarr = self._read_band(
                self.get_default_band_path(),
                band=self.get_default_band(),
                pixel_size=pixel_size,
                size=size,
                as_type=np.uint8,
                masked=False,
                **kwargs,
            )
            mask = def_xarr.copy(data=np.zeros_like(def_xarr.data))

        return mask.rename(self._mask_type.value)

    def open_mask_udm2(
        self,
        mask_id: str,
        pixel_size: float = None,
        size: Union[list, tuple] = None,
        **kwargs,
    ) -> Union[xr.DataArray, None]:
        """
        Open a Planet UDM2 (Usable Data Mask) mask, band by band, as a xarray.
        Returns None if the mask is not available.

        Do not open cloud mask with this function. Use :code:`load` instead.

        See `UDM2 specifications <https://developers.planet.com/docs/data/udm-2/>`_ for more
        information.

        Accepted mask IDs:

        - :code:`CLEAR`:      Band 1     Clear map        [0, 1]  0: not clear, 1: clear
        - :code:`SNOW`:       Band 2     Snow map         [0, 1]  0: no snow or ice, 1: snow or ice
        - :code:`SHADOW`:     Band 3     Shadow map       [0, 1]  0: no shadow, 1: shadow
        - :code:`LIGHT_HAZE`: Band 4     Light haze map   [0, 1]  0: no light haze, 1: light haze
        - :code:`HEAVY_HAZE`: Band 5     Heavy haze map   [0, 1]  0: no heavy haze, 1: heavy haze
        - :code:`CLOUD`:      Band 6     Cloud map        [0, 1]  0: no cloud, 1: cloud
        - :code:`CONFIDENCE`: Band 7     Confidence map   [0-100] %age value: per-pixel algorithmic confidence in classif
        - :code:`UNUSABLE`:   Band 8     Unusable pixels  --      Equivalent to the UDM asset

        .. code-block:: python

            >>> from eoreader.bands import *
            >>> from eoreader.reader import Reader
            >>> path = r"SENTINEL2B_20190401-105726-885_L2A_T31UEQ_D_V2-0.zip"
            >>> prod = Reader().open(path)
            >>> prod.open_mask("EDG", GREEN)
            array([[[0, ..., 0]]], dtype=uint8)

        Args:
            mask_id: Mask ID
            pixel_size (float): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.

        Returns:
            Union[xarray.DataArray, None]: Mask array

        """
        band_mapping = {
            "CLEAR": 1,
            "SNOW": 2,
            "SHADOW": 3,
            "LIGHT_HAZE": 4,
            "HEAVY_HAZE": 5,
            "CLOUD": 6,
            "CONFIDENCE": 7,
            "UNUSABLE": 8,
        }

        assert mask_id in band_mapping
        mask_path = self._get_udm2_path()

        # Open mask band
        mask = utils.read(
            mask_path,
            pixel_size=pixel_size,
            size=size,
            resampling=Resampling.nearest,  # Nearest to keep the flags
            masked=False,
            indexes=[band_mapping[mask_id]],
            **kwargs,
        )

        return mask.astype(np.uint8)

    def open_mask_udm(
        self, pixel_size: float = None, size: Union[list, tuple] = None, **kwargs
    ) -> Union[xr.DataArray, None]:
        """
        Open a Planet UDM (Unusable Data Mask) mask as a xarray.
        For RapidEye, the mask is subsampled to 50m, so this function will interpolate to make it to the correct pixel size
        Returns None if the mask is not available.

        Do not open cloud mask with this function. Use :code:`load` instead.

        See `here <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf/>`_ for more
        information.

        Args:
            pixel_size (float): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.

        Returns:
            Union[xarray.DataArray, None]: Mask array

        """
        mask_path = self._get_udm_path()

        # Open mask band
        return utils.read(
            mask_path,
            pixel_size=pixel_size,
            size=size,
            resampling=Resampling.nearest,  # Nearest to keep the flags
            masked=False,
            as_type=np.uint8,
            **kwargs,
        )

    def _load_nodata(
        self, pixel_size: float = None, size: Union[list, tuple] = None, **kwargs
    ) -> Union[xr.DataArray, None]:
        """
        Load nodata (unimaged pixels) as a numpy array.

        See
        `here <https://assets.planet.com/docs/Planet_Combined_Imagery_Product_Specs_letter_screen.pdf>`_
        (unusable data mask) for more information.

        Args:
            pixel_size (float): Band pixel size in meters
            size (Union[tuple, list]): Size of the array (width, height). Not used if pixel_size is provided.

        Returns:
            Union[xarray.DataArray, None]: Nodata array

        """
        udm = self.open_mask("UNUSABLE", pixel_size, size, **kwargs)
        nodata = udm.copy(data=rasters.read_bit_array(udm.compute(), 0))
        return nodata.rename("NODATA")

    def _get_path(
        self,
        filename: str,
        extension: str,
        invalid_lookahead: Union[str, list] = None,
        as_list=False,
    ) -> Union[list, str]:
        """
        Get either the archived path of the normal path of an asset

        Args:
            filename (str): Filename with wildcards
            extension (str): Extension
            invalid_lookahead (Union[str, list]): Invalid lookahed (string that cannot be placed after the filename)

        Returns:
            Union[list, str]: Paths(s)

        """
        if invalid_lookahead is not None and not isinstance(invalid_lookahead, list):
            invalid_lookahead = [invalid_lookahead]

        ok_paths = []
        try:
            if self.is_archived:
                regex = rf".*{filename}\w*[_]*\.{extension}"

                ok_paths = path.get_archived_rio_path(self.path, regex, as_list=True)
            else:
                ok_paths = [
                    str(p) for p in self.path.glob(f"**/*{filename}*.{extension}")
                ]

            if invalid_lookahead:
                for ok_path in ok_paths.copy():
                    if any(
                        il in path.get_filename(ok_path) for il in invalid_lookahead
                    ):
                        ok_paths.remove(ok_path)

                if not ok_paths:
                    raise FileNotFoundError

            if not as_list:
                ok_paths = ok_paths[0]
        except (FileNotFoundError, IndexError):
            LOGGER.warning(
                f"No file corresponding to *{filename}*.{extension} found in {self.path}"
            )

        return ok_paths

    @cache
    def get_mean_sun_angles(self) -> (float, float):
        """
        Get Mean Sun angles (Azimuth and Zenith angles)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"SENTINEL2A_20190625-105728-756_L2A_T31UEQ_C_V2-2"
            >>> prod = Reader().open(path)
            >>> prod.get_mean_sun_angles()
            (154.554755774838, 27.5941391571236)

        Returns:
            (float, float): Mean Azimuth and Zenith angle
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Open zenith and azimuth angle
        if self.constellation == Constellation.SKY:
            try:
                elev_angle = float(root.findtext(".//sun_elevation"))
                azimuth_angle = float(root.findtext(".//sun_azimuth"))
            except TypeError:
                raise InvalidProductError(
                    "Azimuth or Zenith angles not found in metadata!"
                )
        else:
            try:
                elev_angle = float(
                    root.findtext(f".//{nsmap['opt']}illuminationElevationAngle")
                )
                azimuth_angle = float(
                    root.findtext(f".//{nsmap['opt']}illuminationAzimuthAngle")
                )
            except TypeError:
                raise InvalidProductError(
                    "Azimuth or Zenith angles not found in metadata!"
                )

        # From elevation to zenith
        zenith_angle = 90.0 - elev_angle

        return azimuth_angle, zenith_angle

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Open identifier
        if self.constellation == Constellation.SKY:
            name = root.findtext(".//id")
            if not name:
                raise InvalidProductError("id not found in metadata!")
        else:
            name = root.findtext(f".//{nsmap['eop']}identifier")
            if not name:
                raise InvalidProductError(
                    f"{nsmap['eop']}identifier not found in metadata!"
                )

        return name

    @cache
    def get_mean_viewing_angles(self) -> (float, float, float):
        """
        Get Mean Viewing angles (azimuth, off-nadir and incidence angles)

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_mean_viewing_angles()

        Returns:
            (float, float, float): Mean azimuth, off-nadir and incidence angles
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Open zenith and azimuth angle
        if self.constellation == Constellation.SKY:
            try:
                az = float(root.findtext(".//satellite_azimuth"))
                off_nadir = float(root.findtext(".//view_angle"))
                incidence_angle = None
            except TypeError:
                raise InvalidProductError(
                    "satellite_azimuth or view_angle angles not found in metadata!"
                )
        else:
            try:
                az = float(root.findtext(f".//{nsmap[self._nsmap_key]}azimuthAngle"))
                off_nadir = abs(
                    float(
                        root.findtext(f".//{nsmap[self._nsmap_key]}spaceCraftViewAngle")
                    )
                )
                incidence_angle = float(
                    root.findtext(f".//{nsmap['eop']}incidenceAngle")
                )
            except TypeError:
                raise InvalidProductError(
                    "azimuthAngle, spaceCraftViewAngle or incidenceAngle not found in metadata!"
                )

        return az, off_nadir, incidence_angle

    @cache
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"20210406_015904_37_2407.zip"
            >>> prod = Reader().open(path)
            >>> prod.read_mtd()
            (<Element {http://schemas.planet.com/ps/v1/planet_product_metadata_geocorrected_level}
            EarthObservation at 0x1a2621f03c8>,
            {
                'opt': '{http://earth.esa.int/opt}',
                'gml': '{http://www.opengis.net/gml}',
                'eop': '{http://earth.esa.int/eop}',
                'ps': '{http://schemas.planet.com/ps/v1/planet_product_metadata_geocorrected_level}'
            })

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces as a dict
        """
        try:
            mtd_from_path = "metadata*.xml"
            mtd_archived = r"metadata.*\.xml"

            return self._read_mtd_xml(mtd_from_path, mtd_archived)
        except InvalidProductError:
            # Some RapidEye old product don't have the correct nomenclature
            return self._read_mtd_xml("xml", "xml")

    @cache
    def get_cloud_cover(self) -> float:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_cloud_cover()
            55.5

        Returns:
            float: Cloud cover as given in the metadata
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Get the cloud cover
        if self.constellation == Constellation.SKY:
            try:
                cc = float(root.findtext(".//cloud_percent"))

            except TypeError:
                raise InvalidProductError("cloud_percent not found in metadata!")
        else:
            try:
                cc = float(root.findtext(f".//{nsmap['opt']}cloudCoverPercentage"))

            except TypeError:
                raise InvalidProductError(
                    "opt:cloudCoverPercentage not found in metadata!"
                )

        return cc

    @cache
    def get_orbit_direction(self) -> OrbitDirection:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_orbit_direction().value
            "DESCENDING"

        Returns:
            OrbitDirection: Orbit direction (ASCENDING/DESCENDING)
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()

        # Get the orbit direction
        if self.constellation == Constellation.SKY:
            od = OrbitDirection.DESCENDING
        else:
            try:
                od = OrbitDirection.from_value(
                    root.findtext(f".//{nsmap['eop']}orbitDirection")
                )

            except TypeError:
                raise InvalidProductError("eop:orbitDirection not found in metadata!")

        return od

    def _get_condensed_name(self) -> str:
        """
        Get Planet products condensed name ({date}_{constellation}_{product_type}_{tile}).

        Returns:
            str: Condensed name
        """
        tile = f"_{self.tile_name}" if self.tile_name is not None else ""
        condensed_name = f"{self.get_datetime()}_{self.constellation.name}_{self.product_type.name}{tile}"
        return condensed_name



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
RADARSAT-2 products.
More info `here <https://earth.esa.int/eogateway/documents/20142/0/Radarsat-2-Product-description.pdf/f2783c7b-6a22-cbe4-f4c1-6992f9926dca>`_.
"""
import difflib
import logging
from datetime import datetime
from enum import unique
from typing import Union

import geopandas as gpd
from lxml import etree
from sertit import path, vectors
from sertit.misc import ListEnum
from sertit.vectors import WGS84

from eoreader import DATETIME_FMT, EOREADER_NAME, cache
from eoreader.exceptions import InvalidProductError, InvalidTypeError
from eoreader.products import SarProduct, SarProductType
from eoreader.products.product import OrbitDirection
from eoreader.reader import Reader

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class Rs2ProductType(ListEnum):
    """
    RADARSAT-2 projection identifier.
    Take a look `here <https://earth.esa.int/eogateway/documents/20142/0/Radarsat-2-Product-description.pdf/f2783c7b-6a22-cbe4-f4c1-6992f9926dca>`_.
    """

    SLC = "SLC"
    """Single-look complex"""

    SGX = "SGX"
    """SAR georeferenced extra"""

    SGF = "SGF"
    """SAR georeferenced fine"""

    SSG = "SSG"
    """SAR systematic geocorrected"""

    SPG = "SPG"
    """SAR precision geocorrected"""

    # ScanSar
    SCF = "SCF"
    """ScanSAR fine"""

    SCS = "SCS"
    """ScanSAR sampled"""

    SCN = "SCN"
    """ScanSAR narrow beam. Also a Sensor Mode."""

    SCW = "SCW"
    """ScanSAR wide beam. Also a Sensor Mode."""


@unique
class Rs2SensorMode(ListEnum):
    """
    See here
    `this comparison <https://www.asc-csa.gc.ca/eng/satellites/radarsat/technical-features/radarsat-comparison.asp>`_
    for more information (Beam Modes)

    .. WARNING:: The name in the metadata may vary !
    """

    # Single Beam Modes
    S = "Standard"
    """Standard Mode"""

    W = "Wide"
    """Spotlight Mode"""

    F = "Fine"
    """Wide Mode"""

    WF = "Wide Fine"
    """Wide Fine Mode"""

    MF = "Multi-Look Fine"
    """Multi-Look Fine Mode"""

    WMF = "Wide Multi-Look Fine"
    """Wide Multi-Look Fine Mode"""

    XF = "Extra-Fine"
    """Extra-Fine Mode"""

    U = "Ultra-Fine"
    """Ultra-Fine Mode"""

    WU = "Wide Ultra-Fine"
    """Wide Ultra-Fine Mode"""

    EH = "Extended High"
    """Extended High Mode"""

    EL = "Extended Low"
    """Extended Low Mode"""

    SQ = "Standard Quad-Pol"
    """Standard Quad-Pol Mode"""

    WSQ = "Wide Standard Quad-Pol"
    """Wide Standard Quad-Pol Mode"""

    FQ = "Fine Quad-Pol"
    """Fine Quad-Pol Mode"""

    WFQ = "Wide Fine Quad-Pol"
    """Wide Fine Quad-Pol"""

    # ScanSAR Modes
    SCN = "ScanSAR Narrow"
    """ScanSAR Narrow"""

    SCW = "ScanSAR Wide"
    """ScanSAR Wide"""

    OSVN = "Ocean Surveillance"
    """Ocean Surveillance Mode"""

    DVWF = "Ship Detection"
    """Ship Detection Mode"""

    # Spotlight Mode
    SLA = "Spotlight"
    """Spotlight Mode"""


class Rs2Product(SarProduct):
    """
    Class for RADARSAT-2 Products

    You can use directly the .zip file
    """

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)

        See https://earth.esa.int/eogateway/documents/20142/0/Radarsat-2-Product-description.pdf/f2783c7b-6a22-cbe4-f4c1-6992f9926dca
        - p. 40-41, Table 2-4/5 SGX Product Description
        - p. 43/44, Table 2-6 Single Beam and Spotlight SGF Product Description
        - p. 46, 2-8 Product Description for ScanSAR Narrow and ScanSAR Wide Beam Modes
        - p. 48, 2-11 Product Description for Ship Detection (Detection of Vessels) Beam Mod
        - p. 50, 2-15 Product Description for Ocean Surveillance Beam Mode
        - p. 54/55, 2-19/20 SSG Product Description
        - p. 57/58, 2-22/23 SPG Product Description
        """
        def_pixel_size = -1
        def_res = -1

        # -------------------------------------------------------------
        # Selective Single or Dual Polarization
        # Transmit H and/or V, receive H and/or V
        # F = "Fine", WF = "Wide Fine"
        if self.sensor_mode in [Rs2SensorMode.F, Rs2SensorMode.WF]:
            def_res = 7.7
            if self.product_type in [Rs2ProductType.SGX, Rs2ProductType.SLC]:
                def_pixel_size = 3.125
            elif self.product_type in [
                Rs2ProductType.SGF,
                Rs2ProductType.SSG,
                Rs2ProductType.SPG,
            ]:
                def_pixel_size = 6.25

        # S = "Standard", W = "Wide"
        elif self.sensor_mode in [Rs2SensorMode.S, Rs2SensorMode.W]:
            def_res = 24.7
            if self.product_type in [Rs2ProductType.SGX, Rs2ProductType.SLC]:
                def_pixel_size = 8.0
                # 10 for wide...
            elif self.product_type in [
                Rs2ProductType.SGF,
                Rs2ProductType.SSG,
                Rs2ProductType.SPG,
            ]:
                def_pixel_size = 12.5
        # SCN = "ScanSAR Narrow"
        elif self.sensor_mode == Rs2SensorMode.SCN:
            if self.product_type in [
                Rs2ProductType.SCN,
                Rs2ProductType.SCF,
                Rs2ProductType.SCS,
            ]:
                def_pixel_size = 25.0
                def_res = 50.0
        # SCW = "ScanSAR Wide"
        elif self.sensor_mode == Rs2SensorMode.SCW:
            if self.product_type in [
                Rs2ProductType.SCW,
                Rs2ProductType.SCF,
                Rs2ProductType.SCS,
            ]:
                def_pixel_size = 50.0
                def_res = 100.0

        # -------------------------------------------------------------
        # Polarimetric
        # Transmit H and V on alternate pulses /
        # receive H and V on any pulse
        # FQ = "Fine Quad-Pol", WFQ = "Wide Fine Quad-Pol"
        elif self.sensor_mode in [Rs2SensorMode.FQ, Rs2SensorMode.WFQ]:
            def_res = 7.6
            if self.product_type in [
                Rs2ProductType.SGX,
                Rs2ProductType.SSG,
                Rs2ProductType.SPG,
                Rs2ProductType.SLC,
            ]:
                def_pixel_size = 3.125

        # SQ = "Standard Quad-Pol", "Wide Standard Quad-Pol"
        elif self.sensor_mode in [Rs2SensorMode.SQ, Rs2SensorMode.WSQ]:
            def_res = 7.6
            if self.product_type in [
                Rs2ProductType.SGX,
                Rs2ProductType.SSG,
                Rs2ProductType.SPG,
                Rs2ProductType.SLC,
            ]:
                def_pixel_size = 8.0  # x3.125

        # -------------------------------------------------------------
        # Single Polarization HH
        # Transmit H, receive H
        # EH = "Extended High"
        elif self.sensor_mode == Rs2SensorMode.EH:
            def_res = 24.7
            if self.product_type in [Rs2ProductType.SGX, Rs2ProductType.SLC]:
                def_pixel_size = 8.0
            elif self.product_type in [
                Rs2ProductType.SGF,
                Rs2ProductType.SSG,
                Rs2ProductType.SPG,
            ]:
                def_pixel_size = 12.5

        # EL = "Extended Low"
        elif self.sensor_mode == Rs2SensorMode.EL:
            def_res = 24.7
            if self.product_type in [Rs2ProductType.SGX, Rs2ProductType.SLC]:
                def_pixel_size = 10.0
            elif self.product_type in [
                Rs2ProductType.SGF,
                Rs2ProductType.SSG,
                Rs2ProductType.SPG,
            ]:
                def_pixel_size = 12.5

        # -------------------------------------------------------------
        # Selective Single Polarization
        # Transmit H or V, receive H or V
        # SLA = "Spotlight"
        elif self.sensor_mode == Rs2SensorMode.SLA:
            def_res = 0.8
            if self.product_type in [Rs2ProductType.SGX, Rs2ProductType.SLC]:
                # Range pixel spacing is 1.0 m for incidence angles less than or equal to 48 degrees and 0.8 m for incidence angles greater than 48 degrees.
                def_pixel_size = 0.8
            elif self.product_type in [
                Rs2ProductType.SGF,
                Rs2ProductType.SSG,
                Rs2ProductType.SPG,
            ]:
                def_pixel_size = 0.5

        # U = "Ultra-Fine", WU = "Wide Ultra-Fine"
        elif self.sensor_mode in [Rs2SensorMode.U, Rs2SensorMode.WU]:
            def_res = 2.8
            if self.product_type in [Rs2ProductType.SGX, Rs2ProductType.SLC]:
                # For UF: 1.0 m x 1.0 m for incidence angles less than or equal to 48 degrees. 0.8 m x 0.8 m for incidence angles greater than 48 degrees.
                def_pixel_size = 0.8
            elif self.product_type in [
                Rs2ProductType.SGF,
                Rs2ProductType.SSG,
                Rs2ProductType.SPG,
            ]:
                def_pixel_size = 1.5625

        # XF = "Extra-Fine"
        elif self.sensor_mode == Rs2SensorMode.XF:
            def_res = 4.6
            if self.product_type in [Rs2ProductType.SGX, Rs2ProductType.SLC]:
                # Take 1 look ?
                def_pixel_size = 2.0
                # 4 looks: pix_size = 3.12, res = 7.6
                # 28 looks: pix_size = 5.0, res = 23.5
            if self.product_type in [
                Rs2ProductType.SGF,
                Rs2ProductType.SSG,
                Rs2ProductType.SPG,
            ]:
                # Take 1 look ?
                def_pixel_size = 3.125
                # 4 looks: pix_size = 6.25, res = 7.6
                # 28 looks: pix_size = 8.0, res = 23.5

        # MF = "Multi-Look Fine", WMF = "Wide Multi-Look Fine"
        elif self.sensor_mode in [Rs2SensorMode.MF, Rs2SensorMode.WMF]:
            def_res = 7.6
            if self.product_type in [Rs2ProductType.SGX, Rs2ProductType.SLC]:
                def_pixel_size = 3.125
            elif self.product_type in [
                Rs2ProductType.SGF,
                Rs2ProductType.SSG,
                Rs2ProductType.SPG,
            ]:
                def_pixel_size = 6.25

        # -------------------------------------------------------------
        # Ocean surveillance and detection of vessels
        elif self.sensor_mode == Rs2SensorMode.OSVN:
            def_res = 50.0
            if self.sar_prod_type == SarProductType.CPLX:
                def_pixel_size = 35.0
            else:
                def_pixel_size = 50.0

        elif self.sensor_mode == Rs2SensorMode.DVWF:
            def_res = 35.0
            if self.sar_prod_type == SarProductType.CPLX:
                def_pixel_size = 20.0
            else:
                def_pixel_size = 40.0
        else:
            raise InvalidProductError(f"Unknown sensor mode: {self.sensor_mode}")

        self.pixel_size = def_pixel_size
        self.resolution = def_res

        if self.pixel_size < 0 or self.resolution < 0:
            raise InvalidProductError(
                "There has been an error when setting pixel size or the resolution to your data! Please write an issue on Github."
            )

    def _set_instrument(self) -> None:
        """
        Set instrument

        RADARSAT: https://earth.esa.int/eogateway/missions/radarsat
        """
        self.instrument = "SAR C-band"

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        # Private attributes
        self._raw_band_regex = "*imagery_{}.tif"
        self._band_folder = self.path
        self.snap_filename = ""

        # Its original filename is its name
        self._use_filename = True

        # SNAP can process non-complex archive
        root, nsmap = self.read_mtd()
        namespace = nsmap[None]

        # Open identifier
        prod_type = root.findtext(f".//{namespace}productType")
        if not prod_type:
            raise InvalidProductError("productType not found in metadata!")

        self.product_type = Rs2ProductType.from_value(prod_type)
        self.needs_extraction = self.product_type == Rs2ProductType.SLC

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting product-type, band names and so on)
        """

        # Post init done by the super class
        super()._post_init(**kwargs)

    @cache
    def wgs84_extent(self) -> gpd.GeoDataFrame:
        """
        Get the WGS84 extent of the file before any reprojection.
        This is useful when the SAR pre-process has not been done yet.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"RS2_OK73950_PK661843_DK590667_U25W2_20160228_112418_HH_SGF.zip"
            >>> prod = Reader().open(path)
            >>> prod.wgs84_extent()
                                                        geometry
            1  POLYGON ((106.57999 -6.47363, 107.06926 -6.473...

        Returns:
            gpd.GeoDataFrame: WGS84 extent as a gpd.GeoDataFrame

        """
        # Open extent KML file
        try:
            if self.is_archived:
                product_kml = vectors.read(self.path, archive_regex=r".*product\.kml")
            else:
                extent_file = next(self.path.glob("*product.kml"))
                product_kml = vectors.read(extent_file)
        except IndexError as ex:
            raise InvalidProductError(
                f"Extent file (product.kml) not found in {self.path}"
            ) from ex

        extent_wgs84 = product_kml[
            product_kml.Name == "Polygon Outline"
        ].envelope.to_crs(WGS84)

        return gpd.GeoDataFrame(geometry=extent_wgs84.geometry, crs=extent_wgs84.crs)

    def _set_product_type(self) -> None:
        """Set products type"""
        if self.product_type == Rs2ProductType.SLC:
            self.sar_prod_type = SarProductType.CPLX
        else:
            self.sar_prod_type = SarProductType.GDRG

        if self.product_type not in [Rs2ProductType.SGF, Rs2ProductType.SLC]:
            LOGGER.warning(
                "Other product types than SGF or SLC haven't been tested for %s data. "
                "Use it at your own risks !",
                self.constellation.value,
            )

    def _set_sensor_mode(self) -> None:
        """
        Get products type from RADARSAT-2 products name (could check the metadata too)
        """
        # Get metadata
        root, nsmap = self.read_mtd()
        namespace = nsmap[None]

        # Get sensor mode
        # WARNING: this word may differ from the Enum !!! (no docs available)
        # Get the closest match
        sensor_mode_xml = root.findtext(f".//{namespace}acquisitionType")

        if sensor_mode_xml:
            sensor_mode = difflib.get_close_matches(
                sensor_mode_xml, Rs2SensorMode.list_values()
            )[0]
            try:
                self.sensor_mode = Rs2SensorMode.from_value(sensor_mode)
            except ValueError as ex:
                raise InvalidTypeError(f"Invalid sensor mode for {self.name}") from ex
        else:
            raise InvalidTypeError(f"Invalid sensor mode for {self.name}")

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"RS2_OK73950_PK661843_DK590667_U25W2_20160228_112418_HH_SGF.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2016, 2, 28, 11, 24, 18)
            >>> prod.get_datetime(as_datetime=False)
            '20160228T112418'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        if self.datetime is None:
            # Get MTD XML file
            root, nsmap = self.read_mtd()
            namespace = nsmap[None]

            # Open identifier
            acq_date = root.findtext(f".//{namespace}rawDataStartTime")
            if not acq_date:
                raise InvalidProductError("rawDataStartTime not found in metadata!")

            # Convert to datetime
            date = datetime.strptime(acq_date, "%Y-%m-%dT%H:%M:%S.%fZ")
        else:
            date = self.datetime

        if not as_datetime:
            date = date.strftime(DATETIME_FMT)

        return date

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        name = self.filename

        # Test filename
        reader = Reader()
        if not reader.valid_name(name, self._get_constellation()):
            LOGGER.warning(
                "This RADARSAT-2 filename is not valid. "
                "However RADARSAT-2 files do not provide anywhere the true name of the product. "
                "Use it with caution."
            )

        return name

    @cache
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"LC08_L1GT_023030_20200518_20200527_01_T2"
            >>> prod = Reader().open(path)
            >>> prod.read_mtd()
            (<Element {http://www.rsi.ca/rs2/prod/xml/schemas}product at 0x1c0efbd37c8>,
            {None: '{http://www.rsi.ca/rs2/prod/xml/schemas}'})

        Returns:
            (etree._Element, dict): Metadata XML root and its namespace
        """
        mtd_from_path = "product.xml"
        mtd_archived = r"product\.xml"

        return self._read_mtd_xml(mtd_from_path, mtd_archived)

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing.

        Returns:
            str: Quicklook path
        """
        quicklook_path = None
        try:
            if self.is_archived:
                quicklook_path = path.get_archived_rio_path(
                    self.path, file_regex=r".*BrowseImage\.tif"
                )
            else:
                quicklook_path = str(next(self.path.glob("BrowseImage.tif")))
        except (StopIteration, FileNotFoundError):
            LOGGER.warning(f"No quicklook found in {self.condensed_name}")

        return quicklook_path

    @cache
    def get_orbit_direction(self) -> OrbitDirection:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_orbit_direction().value
            "DESCENDING"

        Returns:
            OrbitDirection: Orbit direction (ASCENDING/DESCENDING)
        """
        # Get MTD XML file
        root, nsmap = self.read_mtd()
        namespace = nsmap[None]

        # Get the orbit direction
        try:
            od = OrbitDirection.from_value(
                root.findtext(f".//{namespace}passDirection").upper()
            )

        except TypeError:
            raise InvalidProductError("passDirection not found in metadata!")

        return od



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Product Factory, class creating products according to their names """

from __future__ import annotations

import importlib
import logging
import re
from enum import unique
from typing import Union
from zipfile import BadZipFile

import pystac
import validators
from pystac import Item
from sertit import AnyPath, path, strings
from sertit.misc import ListEnum
from sertit.types import AnyPathStrType

from eoreader import EOREADER_NAME
from eoreader.exceptions import InvalidProductError

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class CheckMethod(ListEnum):
    """Methods to recognize a product"""

    MTD = "Metadata"
    """Check the metadata: faster method"""

    NAME = "Filename"
    """
    Check the filename:

    Safer method that allows modified product names as it recursively looks for the metadata name in the product files.
    For products that have generic metadata files (ie. RS2 that as mtd named :code:`product.xml`),
    it also checks the band name.
    """

    BOTH = "Both"
    """Check the metadata and the filename: Double check if you have a doubt."""


@unique
class Constellation(ListEnum):
    """Constellation supported by EOReader"""

    S1 = "Sentinel-1"
    """Sentinel-1"""

    S1_RTC_ASF = "Sentinel-1 RTC ASF"
    """
    Sentinel-1 RTC processed by ASF: https://hyp3-docs.asf.alaska.edu/guides/rtc_product_guide/
    """

    S1_RTC_MPC = "Sentinel-1 RTC MPC"
    """
    Sentinel-1 RTC processed by MPC: https://planetarycomputer.microsoft.com/dataset/sentinel-1-rtc
    """

    S2 = "Sentinel-2"
    """Sentinel-2"""

    S2_E84 = "Sentinel-2 stored on AWS and processed by Element84"
    """
    Sentinel-2 stored on AWS and processed by Element84:
    - Element84: arn:aws:s3:::sentinel-cogs - https://registry.opendata.aws/sentinel-2-l2a-cogs

    Not a real constellation, only used for regex.
    """

    S2_MPC = "Sentinel-2 stored on Azure and processed by Microsoft Planetary Computer"
    """
    Sentinel-2 stored on Azure and processed by Microsoft Planetary Computer:
    https://planetarycomputer.microsoft.com/dataset/sentinel-2-l2a

    Not a real constellation, only used for regex.
    """

    S2_SIN = "Sentinel-2 stored on AWS and processed by Sinergise"
    """
    Sentinel-2 stored on AWS and processed by Sinergise:
    arn:aws:s3:::sentinel-s2-l1c and arn:aws:s3:::sentinel-s2-l2a - https://registry.opendata.aws/sentinel-2/

    Not a real constellation, only used for regex.
    """

    S2_THEIA = "Sentinel-2 Theia"
    """Sentinel-2 Theia"""

    S3_OLCI = "Sentinel-3 OLCI"
    """Sentinel-3 OLCI"""

    S3_SLSTR = "Sentinel-3 SLSTR"
    """Sentinel-3 SLSTR"""

    L9 = "Landsat-9"
    """Landsat-9"""

    L8 = "Landsat-8"
    """Landsat-8"""

    L7 = "Landsat-7"
    """Landsat-7"""

    L5 = "Landsat-5"
    """Landsat-5"""

    L4 = "Landsat-4"
    """Landsat-4"""

    L3 = "Landsat-3"
    """Landsat-3"""

    L2 = "Landsat-2"
    """Landsat-2"""

    L1 = "Landsat-1"
    """Landsat-1"""

    PLA = "PlanetScope"
    """PlanetScope"""

    RE = "RapidEye"
    """RapidEye"""

    SKY = "SkySat"
    """SkySat"""

    TSX = "TerraSAR-X"
    """TerraSAR-X"""

    TDX = "TanDEM-X"
    """TanDEM-X"""

    PAZ = "PAZ SAR"
    """SEOSAR/PAZ SAR"""

    RS2 = "RADARSAT-2"
    """RADARSAT-2"""

    PLD = "Pleiades"
    """Pléiades"""

    PNEO = "Pleiades-Neo"
    """Pleiades-Néo"""

    SPOT7 = "Spot-7"
    """SPOT-7"""

    SPOT6 = "Spot-6"
    """SPOT-6"""

    SPOT5 = "Spot-5"
    """SPOT-5"""

    SPOT4 = "Spot-4"
    """SPOT-4"""

    VIS1 = "Vision-1"
    """Vision-1"""

    GS2 = "GEOSAT-2"
    """GEOSAT-2 (ex. DEIMOS-2)"""

    HLS = "HLS"
    """Harmonized Landsat-Sentinel"""

    QB = "QuickBird"
    """QuickBird"""

    GE01 = "GeoEye-1"
    """GeoEye-1"""

    WV01 = "WorldView-1"
    """WorldView-1"""

    WV02 = "WorldView-2"
    """WorldView-2"""

    WV03 = "WorldView-3"
    """WorldView-3"""

    WV04 = "WorldView-4"
    """WorldView-4"""

    RCM = "RADARSAT-Constellation Mission"
    """RADARSAT-Constellation Mission"""

    ICEYE = "ICEYE"
    """ICEYE"""

    SAOCOM = "SAOCOM-1"
    """SAOCOM-1"""

    SV1 = "SuperView-1"
    """SuperView-1"""

    CSK = "COSMO-SkyMed"
    """COSMO-SkyMed"""

    CSG = "COSMO-SkyMed 2nd Generation"
    """COSMO-SkyMed 2nd Generation"""

    CAPELLA = "Capella"
    """Capella"""

    SPOT45 = "Spot-4/5"
    """SPOT-4/5 (not a real constellation, but used as a template for SPOT4/5 products)"""

    MAXAR = "Maxar"
    """Maxar (not a real constellation, but used as a template for every Maxar products)"""

    CUSTOM = "CUSTOM"
    """Custom stack"""


CONSTELLATION_REGEX = {
    Constellation.S1: r"S1[AB]_(IW|EW|SM|WV|S\d)_(RAW|SLC|GRD|OCN)[FHM_]_[0-2]S[SD][HV]_\d{8}T\d{6}_\d{8}T\d{6}_\d{6}_.{11}(_COG|)",
    Constellation.S2: r"S2[AB]_MSIL(1C|2A)_\d{8}T\d{6}_N\d{4}_R\d{3}_T\d{2}\w{3}_\d{8}T\d{6}",
    # Element84 : S2A_31UDQ_20230714_0_L2A, Sinergise: 0 or 1...
    Constellation.S2_E84: r"S2[AB]_\d{2}\w{3}_\d{8}_\d_L(1C|2A)",
    Constellation.S2_THEIA: r"SENTINEL2[AB]_\d{8}-\d{6}-\d{3}_L(2A|1C)_T\d{2}\w{3}_[CDH](_V\d-\d|)",
    Constellation.S3_OLCI: r"S3[AB]_OL_[012]_\w{6}_\d{8}T\d{6}_\d{8}T\d{6}_\d{8}T\d{6}_\w{17}_\w{3}_[OFDR]_(NR|ST|NT)_\d{3}",
    Constellation.S3_SLSTR: r"S3[AB]_SL_[012]_\w{6}_\d{8}T\d{6}_\d{8}T\d{6}_\d{8}T\d{6}_\w{17}_\w{3}_[OFDR]_(NR|ST|NT)_\d{3}",
    Constellation.L9: r"L[OTC]09_(L1(GT|TP)|L2(SP|SR))_\d{6}_\d{8}_\d{8}_\d{2}_(RT|T1|T2)",
    Constellation.L8: r"L[OTC]08_(L1(GT|TP)|L2(SP|SR))_\d{6}_\d{8}_\d{8}_\d{2}_(RT|T1|T2)",
    Constellation.L7: r"LE07_(L1(GT|TP|GS)|L2(SP|SR))_\d{6}_\d{8}_\d{8}_\d{2}_(RT|T1|T2)",
    Constellation.L5: r"L[TM]05_(L1(TP|GS)|L2(SP|SR))_\d{6}_\d{8}_\d{8}_\d{2}_(T1|T2)",
    Constellation.L4: r"L[TM]04_(L1(TP|GS)|L2(SP|SR))_\d{6}_\d{8}_\d{8}_\d{2}_(T1|T2)",
    Constellation.L3: r"LM03_L1(TP|GS)_\d{6}_\d{8}_\d{8}_\d{2}_T2",
    Constellation.L2: r"LM02_L1(TP|GS)_\d{6}_\d{8}_\d{8}_\d{2}_T2",
    Constellation.L1: r"LM01_L1(TP|GS)_\d{6}_\d{8}_\d{8}_\d{2}_T2",
    Constellation.SKY: r"\d{8}_\d{6}_ssc\w{1,4}_\w{4,5}",
    Constellation.PLA: r"\d{8}_\d{6}_(\d{2}_|)\w{4}",
    Constellation.RE: r"(\d{7}_\d{4}-\d{2}-\d{2}_RE\d_3A_\d{6}|\d{4}-\d{2}-\d{2}T\d{6}_RE\d_1B_.+|RE_.+_RE\d_(1B|3A)_.+)",
    Constellation.CSK: [
        r".+",  # Need to check inside as the folder does not have any recognizable name
        r"CSKS\d_(RAW|SCS|DGM|GEC|GTC)_[UB]_(HI|PP|WR|HR|S2)_"
        r"\w{2}_(HH|VV|VH|HV|CO|CH|CV)_[LR][AD]_[FS][NF]_\d{14}_\d{14}\.h5",
    ],
    Constellation.CSG: [
        r".+",  # Need to check inside as the folder does not have any recognizable name
        r"CSG_SSAR\d_(RAW|SCS|DGM|GEC|GTC)_([UBF]|FQLK_B)_\d{4}_(S2[ABC]|D2[RSJ]|OQ[RS]|STR|SC[12]|PPS|QPS)_\d{3}"
        r"_(HH|VV|VH|HV)_[LR][AD]_[DPFR]_\d{14}_\d{14}\_\d_[FC]_\d{2}[NS]_Z\d{2}_[NFB]\d{2}.h5",
    ],
    Constellation.TSX: r"TSX1_SAR__(SSC|MGD|GEC|EEC)_([SR]E|__)___[SH][MCLS]_[SDTQ]_[SD]RA_\d{8}T\d{6}_\d{8}T\d{6}",
    Constellation.TDX: r"TDX1_SAR__(SSC|MGD|GEC|EEC)_([SR]E|__)___[SH][MCLS]_[SDTQ]_[SD]RA_\d{8}T\d{6}_\d{8}T\d{6}",
    Constellation.PAZ: r"PAZ1_SAR__(SSC|MGD|GEC|EEC)_([SR]E|__)___[SH][MCLS]_[SDTQ]_[SD]RA_\d{8}T\d{6}_\d{8}T\d{6}",
    Constellation.RS2: r"RS2_(OK\d+_PK\d+_DK\d+_.{2,}_\d{8}_\d{6}|\d{8}_\d{6}_\d{4}_.{1,5})"
    r"(_(HH|VV|VH|HV)){1,4}_S(LC|GX|GF|CN|CW|CF|CS|SG|PG)(_\d{6}_\d{4}_\d{8}|)",
    Constellation.PLD: r"IMG_PHR1[AB]_(P|MS|PMS|MS-N|MS-X|PMS-N|PMS-X)_\d{3}",
    Constellation.PNEO: r"IMG_\d+_PNEO\d_(P|MS|PMS|MS-FS|PMS-FS)",
    Constellation.SPOT7: r"IMG_SPOT7_(P|MS|PMS|MS-N|MS-X|PMS-N|PMS-X)_\d{3}_\w",
    Constellation.SPOT6: r"IMG_SPOT6_(P|MS|PMS|MS-N|MS-X|PMS-N|PMS-X)_\d{3}_\w",
    Constellation.SPOT45: r"SPVIEW_.+",
    Constellation.SPOT4: r"SP04_HIR_(M_|I_|MI|X_|MX)___\d_\d{8}T\d{6}_\d{8}T\d{6}_.*",
    Constellation.SPOT5: r"SP05_HRG_(HM_|J__|T__|X__|TX__|HMX)__\d_\d{8}T\d{6}_\d{8}T\d{6}_.*",
    Constellation.VIS1: r"VIS1_(PAN|BUN|PSH|MS4)_.+_\d{2}-\d",
    Constellation.RCM: r"RCM\d_OK\d+_PK\d+_\d_.{4,}_\d{8}_\d{6}(_(HH|VV|VH|HV|RV|RH)){1,4}_(SLC|GRC|GRD|GCC|GCD)",
    Constellation.QB: r"\d{12}_\d{2}_P\d{3}_(MUL|PAN|PSH|MOS)",
    Constellation.GE01: r"\d{12}_\d{2}_P\d{3}_(MUL|PAN|PSH|MOS)",
    Constellation.WV01: r"\d{12}_\d{2}_P\d{3}_(MUL|PAN|PSH|MOS)",
    Constellation.WV02: r"\d{12}_\d{2}_P\d{3}_(MUL|PAN|PSH|MOS)",
    Constellation.WV03: r"\d{12}_\d{2}_P\d{3}_(MUL|PAN|PSH|MOS)",
    Constellation.WV04: r"\d{12}_\d{2}_P\d{3}_(MUL|PAN|PSH|MOS)",
    Constellation.MAXAR: r"\d{12}_\d{2}_P\d{3}_(MUL|PAN|PSH|MOS)",
    Constellation.ICEYE: r"((SM|SL|SC|SLEA)[HW]*_\d{5,}|ICEYE_X\d_(SM|SL|SC|SLEA)H*_\d{5,}_\d{8}T\d{6})",
    Constellation.SAOCOM: r".+EOL1[ABCD]SARSAO1[AB]\d+(-product|)",
    Constellation.CAPELLA: r"CAPELLA_C\d{2}_S[PMS]_(GEO|GEC|SLC|SICD|SIDD)_(HH|VV)_\d{14}_\d{14}",
    Constellation.SV1: [
        r"\d{13}_\d{2}",
        r"SV1-0[1-4]_\d{8}_L(1B|2A)\d{10}_\d{13}_\d{2}-(MUX|PSH)\.xml",
    ],
    Constellation.HLS: r"HLS\.[LS]30\.T\d{2}\w{3}\.\d{7}T\d{6}\.v2\.0",
    Constellation.GS2: r"DE2_(PM4|PSH|PS3|PS4|MS4|PAN)_L1[A-D]_\d{6}_\d{8}T\d{6}_\d{8}T\d{6}_DE2_\d{5}_.{4}",
    Constellation.S2_SIN: [r"\d", r"B12\.jp2"],
    Constellation.S1_RTC_ASF: r"S1[AB]_(IW|EW|SM|WV|S\d)_\d{8}T\d{6}_[DS][VH][PRO]_RTC\d{2}_.*",
}

MTD_REGEX = {
    Constellation.S1: {
        "nested": 1,
        # File that can be found at any level (product/**/file)
        "regex": r".*s1[ab]-(iw|ew|sm|wv|s\d)\d*-(raw|slc|grd|ocn)-[hv]{2}-\d{8}t\d{6}-\d{8}t\d{6}-\d{6}-\w{6}-\d{3}(-cog|)\.xml",
    },
    Constellation.S2: {"nested": 3, "regex": r"MTD_TL.xml"},
    Constellation.S2_E84: rf"{CONSTELLATION_REGEX[Constellation.S2_E84]}\.json",
    Constellation.S2_THEIA: rf"{CONSTELLATION_REGEX[Constellation.S2_THEIA]}_MTD_ALL\.xml",
    Constellation.S3_OLCI: r"Oa\d{2}_radiance.nc",
    Constellation.S3_SLSTR: r"S\d_radiance_an.nc",
    Constellation.L9: rf"{CONSTELLATION_REGEX[Constellation.L9]}_MTL\.txt",
    Constellation.L8: rf"{CONSTELLATION_REGEX[Constellation.L8]}_MTL\.txt",
    Constellation.L7: rf"{CONSTELLATION_REGEX[Constellation.L7]}_MTL\.txt",
    Constellation.L5: rf"{CONSTELLATION_REGEX[Constellation.L5]}_MTL\.txt",
    Constellation.L4: rf"{CONSTELLATION_REGEX[Constellation.L4]}_MTL\.txt",
    Constellation.L3: rf"{CONSTELLATION_REGEX[Constellation.L3]}_MTL\.txt",
    Constellation.L2: rf"{CONSTELLATION_REGEX[Constellation.L2]}_MTL\.txt",
    Constellation.L1: rf"{CONSTELLATION_REGEX[Constellation.L1]}_MTL\.txt",
    Constellation.PLA: {
        "nested": -1,  # File that can be found at any level (product/**/file)
        "regex": r"(\d{8}_\d{6}_(\d{2}_|)\w{4}_[13][AB]|\d{7}_\d{7}_\d{4}-\d{2}-\d{2})_.*metadata.*\.xml",
    },
    Constellation.RE: {
        "nested": -1,  # File that can be found at any level (product/**/file)
        "regex": r"(\d{7}_\d{4}-\d{2}-\d{2}_RE\d_3A_\d{7}|\d{4}-\d{2}-\d{2}T\d{6}_RE\d_1B.+)(_.*metadata|).*\.xml",
    },
    Constellation.SKY: {
        "nested": -1,  # File that can be found at any level (product/**/file)
        "regex": r"\d{8}_\d{6}_ssc\w{1,4}_\w{4,5}_.*metadata.*\.json",
    },
    Constellation.CSK: rf"{CONSTELLATION_REGEX[Constellation.CSK][1]}",
    Constellation.CSG: rf"{CONSTELLATION_REGEX[Constellation.CSG][1]}",
    Constellation.TSX: rf"{CONSTELLATION_REGEX[Constellation.TSX]}\.xml",
    Constellation.TDX: rf"{CONSTELLATION_REGEX[Constellation.TDX]}\.xml",
    Constellation.PAZ: rf"{CONSTELLATION_REGEX[Constellation.PAZ]}\.xml",
    Constellation.RS2: [
        r"product\.xml",  # Too generic name, check also a band
        r"imagery_[HV]{2}\.tif",
    ],
    Constellation.PLD: r"DIM_PHR1[AB]_(P|MS|PMS|MS-N|MS-X|PMS-N|PMS-X)_\d{15}_(SEN|PRJ|ORT|MOS)_.{10,}\.XML",
    Constellation.PNEO: r"DIM_PNEO\d_\d{15}_(P|MS|PMS|MS-FS|PMS-FS)_(SEN|PRJ|ORT|MOS)_.{9,}_._._._.\.XML",
    Constellation.SPOT7: r"DIM_SPOT7_(P|MS|PMS|MS-N|MS-X|PMS-N|PMS-X)_\d{15}_(SEN|PRJ|ORT|MOS)_.{10,}\.XML",
    Constellation.SPOT6: r"DIM_SPOT6_(P|MS|PMS|MS-N|MS-X|PMS-N|PMS-X)_\d{15}_(SEN|PRJ|ORT|MOS)_.{10,}\.XML",
    Constellation.VIS1: r"DIM_VIS1_(PSH|MS4|PAN)_\d{14}_(PRJ|ORTP)_S\d{5,}_\d{4}_Meta\.xml",
    Constellation.RCM: {
        "nested": 1,  # File that can be found at 1st folder level (product/*/file)
        "regex": [
            r"product\.xml",  # Too generic name, check also a band
            r"\d+_[RHV]{2}\.tif",
        ],
    },
    Constellation.QB: r"\d{2}\w{3}\d{8}-.{4}(_R\dC\d|)-\d{12}_\d{2}_P\d{3}.TIL",
    Constellation.GE01: r"\d{2}\w{3}\d{8}-.{4}(_R\dC\d|)-\d{12}_\d{2}_P\d{3}.TIL",
    Constellation.WV01: r"\d{2}\w{3}\d{8}-.{4}(_R\dC\d|)-\d{12}_\d{2}_P\d{3}.TIL",
    Constellation.WV02: r"\d{2}\w{3}\d{8}-.{4}(_R\dC\d|)-\d{12}_\d{2}_P\d{3}.TIL",
    Constellation.WV03: r"\d{2}\w{3}\d{8}-.{4}(_R\dC\d|)-\d{12}_\d{2}_P\d{3}.TIL",
    Constellation.WV04: r"\d{2}\w{3}\d{8}-.{4}(_R\dC\d|)-\d{12}_\d{2}_P\d{3}.TIL",
    Constellation.MAXAR: r"\d{2}\w{3}\d{8}-.{4}(_R\dC\d|)-\d{12}_\d{2}_P\d{3}.TIL",
    Constellation.ICEYE: r"ICEYE_(X\d{1,}_|)(SLC|GRD)_((SM|SL|SC)H*|SLEA)_\d{5,}_\d{8}T\d{6}\.xml",
    Constellation.SAOCOM: r"S1[AB]_OPER_SAR_EOSSP__CORE_L1[A-D]_OL(F|VF)_\d{8}T\d{6}.xemt",
    Constellation.CAPELLA: rf"{CONSTELLATION_REGEX[Constellation.CAPELLA]}.*\.json",
    Constellation.SV1: r"SV1-0[1-4]_\d{8}_L(1B|2A)\d{10}_\d{13}_\d{2}-(MUX|PSH)\.xml",
    Constellation.HLS: rf"{CONSTELLATION_REGEX[Constellation.HLS]}\.Fmask\.tif",
    Constellation.GS2: rf"{CONSTELLATION_REGEX[Constellation.GS2]}\.dim",
    Constellation.SPOT45: {
        "nested": -1,  # File that can be found at any level (product/**/file)
        "regex": [
            r"METADATA\.DIM",  # Too generic name, check also a band
            r"IMAGERY\.TIF",
        ],
    },
    Constellation.SPOT4: {
        "nested": -1,  # File that can be found at any level (product/**/file)
        "regex": [
            r"METADATA\.DIM",  # Too generic name, check also a band
            r"IMAGERY\.TIF",
        ],
    },
    Constellation.SPOT5: {
        "nested": -1,  # File that can be found at any level (product/**/file)
        "regex": [
            r"METADATA\.DIM",  # Too generic name, check also a band
            r"IMAGERY\.TIF",
        ],
    },
    Constellation.S2_SIN: {
        "nested": 1,  # File that can be found at any level (product/**/file)
        "regex": [
            r"metadata\.xml",  # Too generic name, check also a band
            r"B12\.jp2",
        ],
    },
    Constellation.S1_RTC_ASF: rf"{CONSTELLATION_REGEX[Constellation.S1_RTC_ASF]}\.kmz",
}


class Reader:
    """
    Factory class creating satellite products according to their names.

    It creates a singleton that you can call only one time per file.
    """

    def __init__(self):
        self._constellation_regex = {}
        self._mtd_regex = {}
        self._mtd_nested = {}

        # Register constellations
        for constellation, regex in CONSTELLATION_REGEX.items():
            self._constellation_regex[constellation] = self._compile(
                regex, prefix="", suffix=""
            )

        # Register metadata
        for constellation, regex in MTD_REGEX.items():
            if isinstance(regex, dict):
                self._mtd_regex[constellation] = self._compile(
                    regex["regex"], prefix=".*", suffix=""
                )
                self._mtd_nested[constellation] = regex["nested"]
            else:
                self._mtd_regex[constellation] = self._compile(
                    regex, prefix=".*", suffix=""
                )
                self._mtd_nested[constellation] = 0

    @staticmethod
    def _compile(regex: Union[str, list], prefix="^", suffix="&") -> list:
        """
        Compile regex or list of regex

        Args:
            regex (Union[str, list]): Regex in :code:`re` sense
            prefix (str): Prefix of regex, ^ by default (means start of the string)
            suffix (str): Prefix of regex, & by default (means end of the string)

        Returns:
            list: List of compiled pattern
        """

        def _compile_(regex_str: str):
            return re.compile(f"{prefix}{regex_str}{suffix}")

        # Case folder is not enough to identify the products (i.e. COSMO Skymed)
        if isinstance(regex, list):
            comp = [_compile_(regex) for regex in regex]
        else:
            comp = [_compile_(regex)]

        return comp

    def open(
        self,
        product_path: AnyPathStrType,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        method: CheckMethod = CheckMethod.MTD,
        remove_tmp: bool = False,
        custom: bool = False,
        constellation: Union[Constellation, str, list] = None,
        **kwargs,
    ) -> "Product":  # noqa: F821
        """
        Open a product from:
        - On disk path
        - Cloud URI (such as s3://)
        - STAC Item URL

        Handled STAC Items are:

        - MPC:
            - S2 L2A COGS
            - Landsat L2 and L1
            - S1 RTC
        - E84:
            - S2 L2A COGS
            - S2 L1C
            - Landsat L2

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2B_MSIL1C_20210517T103619_N7990_R008_T30QVE_20210929T075738.SAFE.zip"
            >>> Reader().open(path)
            eoreader.S2Product 'S2B_MSIL1C_20210517T103619_N7990_R008_T30QVE_20210929T075738'
            Attributes:
                condensed_name: 20210517T103619_S2_T30QVE_L1C_075738
                path: D:/S2B_MSIL1C_20210517T103619_N7990_R008_T30QVE_20210929T075738.SAFE.zip
                constellation: Sentinel-2
                sensor type: Optical
                product type: MSIL1C
                default pixel size: 10.0
                default resolution: 10.0
                acquisition datetime: 2021-05-17T10:36:19
                band mapping:
                    COASTAL_AEROSOL: 01
                    BLUE: 02
                    GREEN: 03
                    RED: 04
                    VEGETATION_RED_EDGE_1: 05
                    VEGETATION_RED_EDGE_2: 06
                    VEGETATION_RED_EDGE_3: 07
                    NIR: 8A
                    NARROW_NIR: 08
                    WATER_VAPOUR: 09
                    CIRRUS: 10
                    SWIR_1: 11
                    SWIR_2: 12
                needs extraction: False
                cloud cover: 0.155752635193646
                tile name: T30QVE

        Args:
            product_path (AnyPathStrType): Product path. URL to a STAC Item is accepted. Cloud URIs (such as s3://...) or classic paths also.
            archive_path (AnyPathStrType): Archive path
            output_path (AnyPathStrType): Output Path
            method (CheckMethod): Checking method used to recognize the products
            remove_tmp (bool): Remove temp files (such as clean or orthorectified bands...) when the product is deleted
            custom (bool): True if we want to use a custom stack
            constellation (Union[Constellation, str, list]): One or several constellations to help the Reader to choose more rapidly the correct Product
            **kwargs: Other arguments
        Returns:
            Product: EOReader's product
        """
        # If an URL is given, it must point to an URL translatable to a STAC Item
        if validators.url(product_path):
            try:
                product_path = pystac.Item.from_file(product_path)
            except Exception:
                raise InvalidProductError(
                    f"Cannot convert your URL ({product_path}) to a STAC Item."
                )

        if isinstance(product_path, Item):
            prod = self._open_stac_item(product_path, output_path, remove_tmp, **kwargs)
        else:
            # If not an Item, it should be a path to somewhere
            prod = self._open_path(
                product_path,
                archive_path,
                output_path,
                method,
                remove_tmp,
                custom,
                constellation,
                **kwargs,
            )

        if not prod:
            LOGGER.warning(
                "There is no existing products in EOReader corresponding to %s",
                product_path,
            )

        return prod

    def _open_stac_item(
        self, item: Item, output_path: AnyPathStrType, remove_tmp: bool, **kwargs
    ):
        """
        Open a STAC Item in EOReader.

        Current STAC handled products: https://github.com/sertit/eoreader/issues/118

        - MPC:
            - S2 L2A COGS
            - Landsat L2 and L1
            - S1 RTC
        - E84:
            - S2 L2A COGS
            - S2 L1C
            - Landsat L2

        Args:
            item (Item): Stac Item
            output_path (AnyPathStrType): Output Path
            remove_tmp (bool): Remove temp files (such as clean or orthorectified bands...) when the product is deleted
            **kwargs: Other arguments

        Returns:
            Product: Product from the STAC Item
        """
        is_mpc = "planetarycomputer" in item.self_href
        is_e84 = "earth-search.aws.element84.com" in item.self_href

        if "rtc" in item.collection_id:
            const = Constellation.S1_RTC_MPC
        else:
            try:
                const = Constellation.from_value(
                    item.properties["constellation"].capitalize().replace(" ", "-")
                )
            except Exception:
                try:
                    const = Constellation.from_value(
                        item.common_metadata.platform.capitalize().replace(" ", "-")
                    )
                except Exception:
                    const = None
                    for const in CONSTELLATION_REGEX.keys():
                        is_valid = self.valid_name(item.id, const)
                        if is_valid:
                            break

        if is_e84 and const == Constellation.S2:
            const = Constellation.S2_E84
        elif is_mpc and const == Constellation.S2:
            const = Constellation.S2_MPC

        if const is not None:
            prod = create_stac_product(
                item=item,
                output_path=output_path,
                remove_tmp=remove_tmp,
                constellation=const,
                is_e84=is_e84,
                is_mpc=is_mpc,
            )
        else:
            prod = None

        return prod

    def _open_path(
        self,
        product_path: AnyPathStrType,
        archive_path: AnyPathStrType = None,
        output_path: AnyPathStrType = None,
        method: CheckMethod = CheckMethod.MTD,
        remove_tmp: bool = False,
        custom: bool = False,
        constellation: Union[Constellation, str, list] = None,
        **kwargs,
    ):
        """
        Open a path or a cloud URI as an EOReader's product

        Args:
            product_path (AnyPathStrType): Product path. URL to a STAC Item is accepted. Cloud URIs (such as s3://...) or classic paths also.
            archive_path (AnyPathStrType): Archive path
            output_path (AnyPathStrType): Output Path
            method (CheckMethod): Checking method used to recognize the products
            remove_tmp (bool): Remove temp files (such as clean or orthorectified bands...) when the product is deleted
            custom (bool): True if we want to use a custom stack
            constellation (Union[Constellation, str, list]): One or several constellations to help the Reader to choose more rapidly the correct Product
            **kwargs: Other arguments
        Returns:
            Product: EOReader's product
        """

        product_path = AnyPath(product_path)

        if not product_path.exists():
            FileNotFoundError(f"Non existing product: {product_path}")

        if custom:
            from eoreader.products import CustomProduct

            prod = CustomProduct(
                product_path=product_path,
                archive_path=archive_path,
                output_path=output_path,
                remove_tmp=remove_tmp,
                constellation=constellation,
                **kwargs,
            )
        else:
            prod = None
            if constellation is None:
                const_list = CONSTELLATION_REGEX.keys()
            else:
                # Manage other products which have the same constellation in them
                if constellation == Constellation.S2:
                    constellation = [
                        Constellation.S2,
                        Constellation.S2_SIN,
                        Constellation.S2_E84,
                        Constellation.S2_MPC,
                    ]
                elif constellation == Constellation.S1:
                    constellation = [
                        Constellation.S1,
                        Constellation.S1_RTC_ASF,
                        Constellation.S1_RTC_MPC,
                    ]

                const_list = Constellation.convert_from(constellation)

            for const in const_list:
                if method == CheckMethod.MTD:
                    is_valid = self.valid_mtd(product_path, const)
                elif method == CheckMethod.NAME:
                    is_valid = self.valid_name(product_path, const)
                else:
                    is_valid = self.valid_name(product_path, const) and self.valid_mtd(
                        product_path, const
                    )

                if is_valid:
                    prod = create_product(
                        product_path=product_path,
                        archive_path=archive_path,
                        output_path=output_path,
                        remove_tmp=remove_tmp,
                        constellation=const,
                        **kwargs,
                    )
                    break

        return prod

    def valid_name(
        self,
        product_path: AnyPathStrType,
        constellation: Union[str, Constellation],
    ) -> bool:
        """
        Check if the product's name is valid for the given satellite


        .. code-block:: python

            >>> from eoreader.reader import Reader, Constellation
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> With IDs
            >>> Reader().valid_name(path, "S1")
            False
            >>> Reader().valid_name(path, "S2")
            True

            >>> # With names
            >>> Reader().valid_name(path, "Sentinel-1")
            False
            >>> Reader().valid_name(path, "Sentinel-2")
            True

            >>> # With Constellation
            >>> Reader().valid_name(path, Constellation.S1)
            False
            >>> Reader().valid_name(path, Constellation.S2)
            True

        Args:
            product_path (AnyPathStrType): Product path
            constellation (str): Constellation's name or ID

        Returns:
            bool: True if valid name

        """
        constellation = Constellation.convert_from(constellation)[0]
        regex = self._constellation_regex[constellation]
        return is_filename_valid(product_path, regex)

    def valid_mtd(
        self,
        product_path: AnyPathStrType,
        constellation: Union[str, Constellation],
    ) -> bool:
        """
        Check if the product's mtd is in the product folder/archive

        .. code-block:: python

            >>> from eoreader.reader import Reader, Constellation
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> With IDs
            >>> Reader().valid_mtd(path, "S1")
            False
            >>> Reader().valid_mtd(path, "S2")
            True

            >>> # With names
            >>> Reader().valid_mtd(path, "Sentinel-1")
            False
            >>> Reader().valid_mtd(path, "Sentinel-2")
            True

            >>> # With Constellation
            >>> Reader().valid_mtd(path, Constellation.S1)
            False
            >>> Reader().valid_mtd(path, Constellation.S2)
            True

        Args:
            product_path (AnyPathStrType): Product path
            constellation (Union[str, Constellation]): Constellation's name or ID

        Returns:
            bool: True if valid name

        """
        # Convert constellation if needed
        constellation = Constellation.convert_from(constellation)[0]

        product_path = AnyPath(product_path)

        if not product_path.exists():
            return False

        # Here the list is a check of several files
        regex_list = self._mtd_regex[constellation]
        nested = self._mtd_nested[constellation]

        # False by default
        is_valid = [False for _ in regex_list]

        # Folder
        if product_path.is_dir():
            if nested < 0:
                prod_files = list(product_path.glob("**/*.*"))
            elif nested == 0:
                prod_files = list(
                    prod_path
                    for prod_path in product_path.iterdir()
                    if prod_path.is_file()
                )
            else:
                nested_wildcard = "/".join(["*" for _ in range(nested)])
                prod_files = list(product_path.glob(f"*{nested_wildcard}/*.*"))

        # Archive
        else:
            try:
                prod_files = path.get_archived_file_list(product_path)
            except BadZipFile:
                raise BadZipFile(f"{product_path} is not a zip file")

        # Check
        for idx, regex in enumerate(regex_list):
            for prod_file in prod_files:
                if regex.match(str(prod_file)):
                    is_valid[idx] = True
                    break

        return all(is_valid)


def is_filename_valid(
    product_path: AnyPathStrType, regex: Union[list, re.Pattern]
) -> bool:
    """
    Check if the filename corresponds to the given satellite regex.

    Checks also if a file inside the directory is correct.

    .. WARNING::
        Two levels maximum for the moment

    Args:
        product_path (AnyPathStrType): Product path
        regex (Union[list, re.Pattern]): Regex or list of regex

    Returns:
        bool: True if the filename corresponds to the given satellite regex
    """
    product_path = AnyPath(product_path)
    # Handle HLS folders...
    if product_path.is_dir() and product_path.suffix in [".0"]:
        product_file_name = product_path.name
    else:
        product_file_name = path.get_filename(product_path)

    # Case folder is not enough to identify the products (i.e. COSMO Skymed)
    # WARNING: Two level max for the moment
    is_valid = bool(regex[0].match(product_file_name))
    if is_valid and len(regex) > 1:
        is_valid = False  # Reset
        if product_path.is_dir():
            file_list = product_path.iterdir()
            for file in file_list:
                if regex[1].match(file.name):
                    is_valid = True
                    break
        else:
            try:
                file_list = path.get_archived_file_list(product_path)
                for file in file_list:
                    if regex[1].match(file):
                        is_valid = True
                        break
            except TypeError:
                LOGGER.debug(
                    f"The product {product_file_name} should be a folder or an archive (.tar or .zip)"
                )
            except BadZipFile:
                raise BadZipFile(f"{product_path} is not a zip file")

            except FileNotFoundError:
                pass

    return is_valid


def create_product(
    product_path: AnyPathStrType,
    archive_path: AnyPathStrType,
    output_path: AnyPathStrType,
    remove_tmp: bool,
    constellation: Constellation,
    **kwargs,
):
    """
    Create Product

    Args:
        product_path (AnyPathStrType): Product path
        archive_path (AnyPathStrType): Archive path
        output_path (AnyPathStrType): Output path
        remove_tmp (bool): Remove tmp files
        constellation (Constellation): COnstellation
        **kwargs: Other arguments

    Returns:
        Product: EOReader product
    """
    sat_class = constellation.name.lower() + "_product"

    # Channel correctly the constellations to their generic files (just in case)
    # TerraSAR-like constellations
    if constellation in [Constellation.TDX, Constellation.PAZ]:
        sat_class = "tsx_product"
        constellation = None  # All product names are the same, so assess it with MTD
    # Maxar-like constellations
    elif constellation in [
        Constellation.QB,
        Constellation.GE01,
        Constellation.WV01,
        Constellation.WV02,
        Constellation.WV03,
        Constellation.WV04,
    ]:
        sat_class = "maxar_product"
        constellation = None  # All product names are the same, so assess it with MTD
    # Lansat constellations
    elif constellation in [
        Constellation.L1,
        Constellation.L2,
        Constellation.L3,
        Constellation.L4,
        Constellation.L5,
        Constellation.L7,
        Constellation.L8,
        Constellation.L9,
    ]:
        sat_class = "landsat_product"
    # SPOT-6/7 constellations
    elif constellation in [Constellation.SPOT6, Constellation.SPOT7]:
        sat_class = "spot67_product"
    # SPOT-4/5 constellations
    elif constellation in [Constellation.SPOT4, Constellation.SPOT5]:
        sat_class = "spot45_product"
    elif constellation in [Constellation.S2_SIN]:
        sat_class = "s2_product"
        kwargs["is_sinergise"] = True

    # Manage both optical and SAR
    try:
        mod = importlib.import_module(f"eoreader.products.sar.{sat_class}")
    except ModuleNotFoundError:
        mod = importlib.import_module(f"eoreader.products.optical.{sat_class}")

    # Get class
    class_ = getattr(mod, strings.snake_to_camel_case(sat_class))

    # Create product
    prod = class_(
        product_path=product_path,
        archive_path=archive_path,
        output_path=output_path,
        remove_tmp=remove_tmp,
        **kwargs,
    )
    return prod


def create_stac_product(
    item: Item,
    output_path: AnyPathStrType,
    remove_tmp: bool,
    constellation: Constellation,
    **kwargs,
):
    """
    Create STAC Product

    Args:
        item (Item): Product path
        output_path (AnyPathStrType): Output path
        remove_tmp (bool): Remove tmp files
        constellation (Constellation): COnstellation
        **kwargs: Other arguments

    Returns:
        Product: EOReader product
    """
    if constellation in [
        Constellation.L1,
        Constellation.L2,
        Constellation.L3,
        Constellation.L4,
        Constellation.L5,
        Constellation.L7,
        Constellation.L8,
        Constellation.L9,
    ]:
        sat_class = "landsat_product"
    else:
        sat_class = constellation.name.lower() + "_product"

    # Manage both optical and SAR
    try:
        mod = importlib.import_module(f"eoreader.products.sar.{sat_class}")
    except ModuleNotFoundError:
        mod = importlib.import_module(f"eoreader.products.optical.{sat_class}")

    # Get class
    if kwargs.get("is_mpc", False) and "mpc" not in sat_class:
        replacement = "MpcStacProduct"
    else:
        replacement = "StacProduct"

    class_name = strings.snake_to_camel_case(sat_class).replace("Product", replacement)
    class_ = getattr(mod, class_name)

    # Create product
    prod = class_(
        item=item,
        remove_tmp=remove_tmp,
        **kwargs,
    )
    return prod



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
SAR and Optical products
"""
# flake8: noqa
__all__ = ["Product", "SensorType", "OrbitDirection"]

from .product import Product, SensorType, OrbitDirection

__all__ += [
    "CustomProduct",
    "CustomFields",
]
from .custom_product import CustomProduct, CustomFields

# STAC products
__all__ += [
    "StacProduct",
]
from .stac_product import StacProduct

# -- Optical --
__all__ += [
    "OpticalProduct",
    "CleanMethod",
]
from .optical.optical_product import OpticalProduct, CleanMethod

# VHR
__all__ += [
    "VhrProduct",
    "DimapV1Product",
    "DimapV2Product",
    "DimapV2BandCombination",
    "DimapV2ProductType",
    "PldProduct",
    "Spot67Product",
    "MaxarProduct",
    "MaxarProductType",
    "MaxarSatId",
    "MaxarBandId",
    "Vis1Product",
    "Vis1ProductType",
    "Vis1BandCombination",
    "Sv1Product",
    "Sv1ProductType",
    "Sv1BandCombination",
    "Gs2Product",
    "Gs2ProductType",
    "Gs2BandCombination",
]
from .optical.vhr_product import VhrProduct
from .optical.dimap_v1_product import DimapV1Product
from .optical.dimap_v2_product import (
    DimapV2BandCombination,
    DimapV2Product,
    DimapV2ProductType,
)
from .optical.pld_product import PldProduct
from .optical.spot67_product import Spot67Product
from .optical.maxar_product import (
    MaxarProduct,
    MaxarProductType,
    MaxarSatId,
    MaxarBandId,
)
from .optical.vis1_product import Vis1Product, Vis1ProductType, Vis1BandCombination
from .optical.sv1_product import Sv1Product, Sv1ProductType, Sv1BandCombination
from .optical.gs2_product import Gs2Product, Gs2ProductType, Gs2BandCombination

# SPOT4/5
__all__ += [
    "Spot45ProductType",
    "Spot4BandCombination",
    "Spot5BandCombination",
    "Spot45Product",
]
from .optical.spot45_product import (
    Spot45ProductType,
    Spot4BandCombination,
    Spot5BandCombination,
    Spot45Product,
)

# Planet
__all__ += [
    "PlanetMaskType",
]

from .optical.planet_product import PlanetMaskType

# PlanetScope
__all__ += [
    "PlaProduct",
    "PlaProductType",
    "PlaInstrument",
]
from .optical.pla_product import PlaProduct, PlaProductType, PlaInstrument

# SkySat
__all__ += [
    "SkyProductType",
    "SkyProduct",
    "SkyInstrument",
]
from .optical.sky_product import SkyProductType, SkyProduct, SkyInstrument

# RapidEye
__all__ += [
    "ReProductType",
    "ReProduct",
    "ReInstrument",
]
from .optical.re_product import ReProductType, ReProduct

# Landsat
__all__ += [
    "LandsatProduct",
    "LandsatProductType",
    "LandsatCollection",
    "LandsatInstrument",
]
from .optical.landsat_product import (
    LandsatProduct,
    LandsatProductType,
    LandsatCollection,
    LandsatInstrument,
)

# Sentinel
__all__ += [
    "S2Product",
    "S2ProductType",
    "S2GmlMasks",
    "S2Jp2Masks",
    "S2StacProduct",
    "S2E84Product",
    "S2E84StacProduct",
    "S2MPCStacProduct",
    "S2TheiaProduct",
    "S3Product",
    "S3ProductType",
    "S3DataType",
    "S3Instrument",
    "S3SlstrProduct",
    "SlstrRadAdjustTuple",
    "SlstrRadAdjust",
    "SlstrView",
    "SlstrStripe",
]
from .optical.s2_product import (
    S2Product,
    S2ProductType,
    S2GmlMasks,
    S2Jp2Masks,
    S2StacProduct,
)
from .optical.s2_e84_product import S2E84Product, S2E84StacProduct
from .optical.s2_mpc_product import S2MpcStacProduct
from .optical.s2_theia_product import S2TheiaProduct
from .optical.s3_product import S3Product, S3ProductType, S3DataType, S3Instrument
from .optical.s3_olci_product import S3OlciProduct
from .optical.s3_slstr_product import (
    S3SlstrProduct,
    SlstrRadAdjustTuple,
    SlstrRadAdjust,
    SlstrView,
    SlstrStripe,
)

# -- SAR --
__all__ += [
    "SarProduct",
    "SarProductType",
    "SnapDems",
    "CosmoProduct",
    "CosmoProductType",
    "CsgProduct",
    "CsgSensorMode",
    "CskProduct",
    "CskSensorMode",
    "IceyeProduct",
    "IceyeProductType",
    "IceyeSensorMode",
    "RcmProduct",
    "RcmProductType",
    "RcmSensorMode",
    "Rs2Product",
    "Rs2ProductType",
    "Rs2SensorMode",
    "S1Product",
    "S1SensorMode",
    "S1ProductType",
    "S1RtcAsfProduct",
    "S1RtcProductType",
    "S1RtcMpcStacProduct",
    "SaocomProduct",
    "SaocomProductType",
    "SaocomPolarization",
    "SaocomSensorMode" "TsxProduct",
    "TsxPolarization",
    "TsxSatId",
    "TsxProductType",
    "TsxSensorMode",
    "CapellaProduct",
    "CapellaProductType",
    "CapellaSensorMode" "",
]
from .sar.sar_product import SarProduct, SarProductType, SnapDems
from .sar.cosmo_product import CosmoProduct, CosmoProductType
from .sar.csg_product import CsgProduct, CsgSensorMode
from .sar.csk_product import CskProduct, CskSensorMode
from .sar.iceye_product import IceyeProduct, IceyeProductType, IceyeSensorMode
from .sar.rcm_product import RcmProduct, RcmProductType, RcmSensorMode
from .sar.rs2_product import Rs2Product, Rs2ProductType, Rs2SensorMode
from .sar.s1_product import S1Product, S1SensorMode, S1ProductType
from .sar.s1_rtc_asf_product import S1RtcAsfProduct, S1RtcProductType
from .sar.s1_rtc_mpc_product import S1RtcMpcStacProduct
from .sar.saocom_product import (
    SaocomProduct,
    SaocomProductType,
    SaocomPolarization,
    SaocomSensorMode,
)
from .sar.tsx_product import (
    TsxProduct,
    TsxPolarization,
    TsxSatId,
    TsxProductType,
    TsxSensorMode,
)
from .sar.capella_product import (
    CapellaProduct,
    CapellaProductType,
    CapellaSensorMode,
)



# -*- coding: utf-8 -*-
# Copyright 2024, SERTIT-ICube - France, https://sertit.unistra.fr/
# This file is part of eoreader project
#     https://github.com/sertit/eoreader
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Sentinel-1 products """
import logging
import os
import re
import tempfile
import zipfile
from datetime import datetime
from enum import unique
from typing import Union

import geopandas as gpd
from lxml import etree
from sertit import path, vectors
from sertit.misc import ListEnum

from eoreader import DATETIME_FMT, EOREADER_NAME, cache
from eoreader.exceptions import InvalidProductError
from eoreader.products import SarProduct, SarProductType
from eoreader.products.product import OrbitDirection

LOGGER = logging.getLogger(EOREADER_NAME)


@unique
class S1ProductType(ListEnum):
    """
    S1 products types. Take a look here:
    https://sentinel.esa.int/web/sentinel/missions/sentinel-1/data-products
    """

    RAW = "RAW"
    """Raw products (lvl 0): **not used by EOReader**"""

    SLC = "SLC"
    """Single Look Complex (SLC, lvl 1)"""

    GRD = "GRD"
    """Ground Range Detected (GRD, lvl 1, phase lost)"""

    OCN = "OCN"
    """Ocean products (lvl 2): **not used by EOReader**"""


@unique
class S1SensorMode(ListEnum):
    """
    S1 sensor mode. Take a look here:
    https://sentinel.esa.int/web/sentinel/user-guides/sentinel-1-sar/acquisition-modes

    The primary conflict-free modes are IW, with VV+VH polarisation over land,
    and WV, with VV polarisation, over open ocean.
    EW mode is primarily used for wide area coastal monitoring including ship traffic, oil spill and sea-ice monitoring.
    SM mode is only used for small islands and on request for extraordinary events such as emergency management.
    """

    SM = "SM"
    """Stripmap (SM)"""

    IW = "IW"
    """Interferometric Wide swath (IW)"""

    EW = "EW"
    """Extra-Wide swath (EW)"""

    WV = "WV"
    """Wave (WV) -> single polarisation only (HH or VV)"""


@unique
class S1ResolutionClass(ListEnum):
    """
    S1 resolution class:
    https://sentinel.esa.int/web/sentinel/user-guides/sentinel-1-sar/resolutions/level-1-ground-range-detected
    """

    FR = "F"
    """Full Resolution (FR)"""

    HR = "H"
    """High Resolution (HR)"""

    MR = "M"
    """Medium Resolution (MR)"""

    NONE = "_"
    """No specified resolution for other product type than GRD"""


class S1Product(SarProduct):
    """
    Class for Sentinel-1 Products

    You can use directly the .zip file
    """

    def _set_pixel_size(self) -> None:
        """
        Set product default pixel size (in meters)
        See here
        `here <https://sentinel.esa.int/web/sentinel/user-guides/sentinel-1-sar/resolutions/level-1-ground-range-detected>`_
        for more information
        """

        if self.product_type == S1ProductType.GRD:
            res_class = S1ResolutionClass.from_value(self.split_name[2][-1])
        else:
            res_class = S1ResolutionClass.NONE

        # Using the az resolution between rg and az
        default_res = {
            S1SensorMode.SM: {
                S1ResolutionClass.NONE: 9.0,
                S1ResolutionClass.FR: 9.0,
                S1ResolutionClass.HR: 23.0,
                S1ResolutionClass.MR: 84.0,
            },
            S1SensorMode.IW: {
                S1ResolutionClass.NONE: 20.0,
                S1ResolutionClass.HR: 20.0,
                S1ResolutionClass.MR: 87.0,
            },
            S1SensorMode.EW: {
                S1ResolutionClass.NONE: 50.0,
                S1ResolutionClass.HR: 50.0,
                S1ResolutionClass.MR: 87.0,
            },
            S1SensorMode.WV: {
                S1ResolutionClass.NONE: 51.0,
                S1ResolutionClass.MR: 51.0,
            },
        }

        # Pixel sizes are squared -> no issue between rg and az
        default_pix_size = {
            S1SensorMode.SM: {
                S1ResolutionClass.NONE: 3.5,
                S1ResolutionClass.FR: 3.5,
                S1ResolutionClass.HR: 10.0,
                S1ResolutionClass.MR: 40.0,
            },
            S1SensorMode.IW: {
                S1ResolutionClass.NONE: 20.0,
                S1ResolutionClass.HR: 20.0,
                S1ResolutionClass.MR: 40.0,
            },
            S1SensorMode.EW: {
                S1ResolutionClass.NONE: 25.0,
                S1ResolutionClass.HR: 25.0,
                S1ResolutionClass.MR: 40.0,
            },
            S1SensorMode.WV: {
                S1ResolutionClass.NONE: 25.0,
                S1ResolutionClass.MR: 25.0,
            },
        }

        try:
            def_pixel_size = default_pix_size[self.sensor_mode][res_class]
            def_res = default_res[self.sensor_mode][res_class]
        except KeyError:
            raise InvalidProductError(f"Unknown sensor mode: {self.sensor_mode}")

        self.pixel_size = def_pixel_size
        self.resolution = def_res

    def _set_instrument(self) -> None:
        """
        Set instrument

        Sentinel-1: https://sentinels.copernicus.eu/web/sentinel/missions/sentinel-1/instrument-payload
        """
        self.instrument = "SAR C-band"

    def _pre_init(self, **kwargs) -> None:
        """
        Function used to pre_init the products
        (setting needs_extraction and so on)
        """
        # Private attributes
        self._raw_band_regex = "*-{!l}-*.tiff"  # Just get the SLC-iw1 image for now
        self._band_folder = self.path.joinpath("measurement")
        self.snap_filename = ""

        # Its original filename is its name
        self._use_filename = True

        # Check if COG in name
        if "_COG" in self.filename:
            raise NotImplementedError(
                "These S1 COG products are not yet handled by SNAP. "
                "EOReader will handle them when this issue is fixed. "
                "See https://forum.step.esa.int/t/handle-sentinel-1-cog-collection/40840. "
                "Please use the classical format instead."
            )

        # Zipped and SNAP can process its archive
        self.needs_extraction = False

        # Pre init done by the super class
        super()._pre_init(**kwargs)

    def _post_init(self, **kwargs) -> None:
        """
        Function used to post_init the products
        (setting product-type, band names and so on)
        """

        # Post init done by the super class
        super()._post_init(**kwargs)

    @cache
    def wgs84_extent(self) -> gpd.GeoDataFrame:
        """
        Get the WGS84 extent of the file before any reprojection.
        This is useful when the SAR pre-process has not been done yet.

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S1A_IW_GRDH_1SDV_20191215T060906_20191215T060931_030355_0378F7_3696.zip"
            >>> prod = Reader().open(path)
            >>> prod.wgs84_extent()
                                   Name  ...                                           geometry
            0  Sentinel-1 Image Overlay  ...  POLYGON ((0.85336 42.24660, -2.32032 42.65493,...
            [1 rows x 12 columns]

        Returns:
            gpd.GeoDataFrame: WGS84 extent as a gpd.GeoDataFrame

        """
        tmp_dir = tempfile.TemporaryDirectory()

        try:
            # Open the map-overlay file
            if self.is_archived:
                # We need to extract the file here as we need a proper file
                with zipfile.ZipFile(self.path, "r") as zip_ds:
                    filenames = [f.filename for f in zip_ds.filelist]
                    regex = re.compile(".*preview.*map-overlay.kml")
                    preview_overlay = zip_ds.extract(
                        list(filter(regex.match, filenames))[0], tmp_dir.name
                    )
            else:
                preview_overlay = self.path.joinpath("preview", "map-overlay.kml")

            if os.path.isfile(preview_overlay):
                # Open the KML file
                extent_wgs84 = vectors.read(preview_overlay)
                if extent_wgs84.empty:
                    raise InvalidProductError(
                        f"Cannot determine the WGS84 extent of {self.name}"
                    )
            else:
                raise InvalidProductError(
                    f"Impossible to find the map-overlay.kml in {self.path}"
                )

        except Exception as ex:
            raise InvalidProductError(ex) from ex

        finally:
            tmp_dir.cleanup()

        return extent_wgs84

    def _set_product_type(self) -> None:
        """Set products type"""
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open identifier
        prod_type = root.findtext(".//productType")
        if not prod_type:
            raise InvalidProductError("mode not found in metadata!")

        self.product_type = S1ProductType.from_value(prod_type)

        if self.product_type == S1ProductType.GRD:
            self.sar_prod_type = SarProductType.GDRG
        elif self.product_type == S1ProductType.SLC:
            self.sar_prod_type = SarProductType.CPLX
        else:
            raise NotImplementedError(
                f"{self.product_type.value} product type is not available for {self.name}"
            )

    def _set_sensor_mode(self) -> None:
        """
        Get products type from S1 products name (could check the metadata too)
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Open identifier
        mode = root.findtext(".//mode")
        if not mode:
            raise InvalidProductError("mode not found in metadata!")

        # Mono swath SM
        if mode in ["S1", "S2", "S3", "S4", "S5", "S6"]:
            mode = "SM"

        # Get sensor mode
        self.sensor_mode = S1SensorMode.from_value(mode)

        if not self.sensor_mode:
            raise InvalidProductError(
                f"Invalid {self.constellation.value} name: {self.name}"
            )

    def get_datetime(self, as_datetime: bool = False) -> Union[str, datetime]:
        """
        Get the product's acquisition datetime, with format :code:`YYYYMMDDTHHMMSS` <-> :code:`%Y%m%dT%H%M%S`

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S1A_IW_GRDH_1SDV_20191215T060906_20191215T060931_030355_0378F7_3696.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_datetime(as_datetime=True)
            datetime.datetime(2019, 12, 15, 6, 9, 6)
            >>> prod.get_datetime(as_datetime=False)
            '20191215T060906'

        Args:
            as_datetime (bool): Return the date as a datetime.datetime. If false, returns a string.

        Returns:
             Union[str, datetime.datetime]: Its acquisition datetime
        """
        if self.datetime is None:
            # Get MTD XML file
            root, _ = self.read_mtd()

            # Open identifier
            acq_date = root.findtext(".//startTime")
            if not acq_date:
                raise InvalidProductError("startTime not found in metadata!")

            # Convert to datetime
            date = datetime.strptime(acq_date, "%Y-%m-%dT%H:%M:%S.%f")
        else:
            date = self.datetime

        if not as_datetime:
            date = date.strftime(DATETIME_FMT)

        return date

    def _get_name_constellation_specific(self) -> str:
        """
        Set product real name from metadata

        Returns:
            str: True name of the product (from metadata)
        """
        try:
            if self.is_archived:
                pdf_file = path.get_archived_path(self.path, r".*\.pdf", as_list=False)
            else:
                pdf_file = next(self.path.glob("*.pdf"))
        except (FileNotFoundError, StopIteration):
            # The name is not in the classic metadata, but can be found in the product-preview
            try:
                mtd_from_path = "preview/product-preview.html"
                mtd_archived = r"preview.*product-preview\.html"

                root = self._read_mtd_html(mtd_from_path, mtd_archived)

                # Open identifier
                name = root.findtext(".//head/title")
                if not name:
                    raise InvalidProductError("title not found in metadata!")

                LOGGER.warning(
                    "Product filename is not a valid Sentinel-1 name, and the retrieved name is missing the unique ID."
                )

            except InvalidProductError:
                raise InvalidProductError(
                    "product-preview.html not found in the product, the name will be the filename (which is not a valid Sentinel-1 name)"
                )
        else:
            name = path.get_filename(pdf_file)

        return name

    @cache
    def _read_mtd(self) -> (etree._Element, dict):
        """
        Read metadata and outputs the metadata XML root and its namespaces as a dict

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S1A_IW_GRDH_1SDV_20191215T060906_20191215T060931_030355_0378F7_3696.zip"
            >>> prod = Reader().open(path)
            >>> prod.read_mtd()
            (<Element product at 0x1832895d788>, {})

        Returns:
            (etree._Element, dict): Metadata XML root and its namespaces
        """
        mtd_from_path = "annotation/*.xml"

        # When archived, other XML (in calibration folder) can be found
        mtd_archived = r"annotation/(?!rfi)(?!cal)(?!noise).*\.xml"

        return self._read_mtd_xml(mtd_from_path, mtd_archived)

    def get_quicklook_path(self) -> str:
        """
        Get quicklook path if existing.

        Returns:
            str: Quicklook path
        """
        quicklook_path = None
        try:
            if self.is_archived:
                quicklook_path = self.path / path.get_archived_path(
                    self.path, file_regex=r".*preview.quick-look\.png"
                )
            else:
                quicklook_path = next(self.path.glob("preview/quick-look.png"))
            quicklook_path = str(quicklook_path)
        except (StopIteration, FileNotFoundError):
            LOGGER.warning(f"No quicklook found in {self.condensed_name}")

        return quicklook_path

    @cache
    def get_orbit_direction(self) -> OrbitDirection:
        """
        Get cloud cover as given in the metadata

        .. code-block:: python

            >>> from eoreader.reader import Reader
            >>> path = r"S2A_MSIL1C_20200824T110631_N0209_R137_T30TTK_20200824T150432.SAFE.zip"
            >>> prod = Reader().open(path)
            >>> prod.get_orbit_direction().value
            "DESCENDING"

        Returns:
            OrbitDirection: Orbit direction (ASCENDING/DESCENDING)
        """
        # Get MTD XML file
        root, _ = self.read_mtd()

        # Get the orbit direction
        try:
            od = OrbitDirection.from_value(root.findtext(".//pass").upper())

        except TypeError:
            raise InvalidProductError("pass not found in metadata!")

        return od
